/*
 *
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.google.gson;

import java.lang.reflect.Field;
import java.util.Map;

import com.google.gson.util.N;


public interface FieldBindingStrategy {
    public static final FieldBindingStrategy GET_METHOD = new FieldBindingStrategy() {
            @Override
            public boolean matches(Field field) {
                return N.getPropGetMethod(field.getDeclaringClass(), field.getName()) != null;
            }
        };

    boolean matches(Field field);

    public static class FieldModifier implements FieldBindingStrategy {
        private static final Map<Integer, FieldModifier> pool = N.asConcurrentHashMap();
        private final int mod;

        public FieldModifier(int mod) {
            this.mod = mod;
        }

        public static FieldModifier valueOf(int mod) {
            FieldModifier result = pool.get(mod);

            if (result == null) {
                result = new FieldModifier(mod);
                pool.put(mod, result);
            }

            return result;
        }

        @Override
        public boolean matches(Field field) {
            return (mod & field.getModifiers()) != 0;
        }
    }
}
