package com.google.gson.util;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import javax.xml.datatype.XMLGregorianCalendar;


/**
 * It's not multi-thread safety.
 * @author haiyangl
 *
 */
public class BufferedWriter extends Writer {
    protected static final int DEFAULT_BUFFER_SIZE = 8192;
    protected static final int DEFAULT_ARRAY_SIZE = 8192;
    protected static final List<char[]> bufferPool = N.newArrayList(100);
    protected static final List<char[]> arrayPool = N.newArrayList(100);
    protected char[] cArray;
    protected char[] cbuff;
    protected int nextChar = 0;
    protected Writer out;
    protected StringBuilder sb;
    private boolean isStringBuilderRecycable = false;

    public BufferedWriter(StringBuilder sb) {
        this.out = null;
        this.sb = sb;
        lock = sb;
    }

    public BufferedWriter(OutputStream os) {
        this(new OutputStreamWriter(os, N.UTF_8));
    }

    public BufferedWriter(Writer writer) {
        this.sb = null;
        this.out = writer;
        lock = writer;
    }

    public void write(char c) throws IOException {
        //        synchronized (lock) {
        if (nextChar >= DEFAULT_BUFFER_SIZE) {
            flushBuffer();
        }

        if (cbuff == null) {
            cbuff = createBuffer();
        }

        cbuff[nextChar++] = c;

        //        }
    }

    public void write(boolean b) throws IOException {
        write(b ? N.TRUE_CHAR_ARRAY : N.FALSE_CHAR_ARRAY);
    }

    public void write(byte b) throws IOException {
        if (b < 0) {
            write(IOUtil.cbufOfInt[-b]);
        } else {
            write(IOUtil.cbufOfInt[b]);
        }
    }

    public void write(short s) throws IOException {
        if (s < 0) {
            write(IOUtil.cbufOfInt[-s]);
        } else {
            write(IOUtil.cbufOfInt[s]);
        }
    }

    /**
     * @deprecated
     * Replaced with write(char).
     */
    @Override
    public void write(int ch) throws IOException {
        write((char) ch);
    }

    public void writeInt(int i) throws IOException {
        if (i < IOUtil.maxCBufIntValue) {
            if (i < 0) {
                if (i == Integer.MIN_VALUE) {
                    write(IOUtil._2147483648_CHAR_ARRAY);

                    return;
                } else if (-i < IOUtil.maxCBufIntValue) {
                    write('-');
                    write(IOUtil.cbufOfInt[-i]);

                    return;
                }
            } else {
                write(IOUtil.cbufOfInt[i]);

                return;
            }
        }

        write(String.valueOf(i));
    }

    public void write(long l) throws IOException {
        if (l < IOUtil.maxCBufIntValue) {
            if (l < 0) {
                if (l == Long.MIN_VALUE) {
                    write(IOUtil._9223372036854775808_CHAR_ARRAY);

                    return;
                } else if (-l < IOUtil.maxCBufIntValue) {
                    write('-');
                    write(IOUtil.cbufOfInt[(int) -l]);

                    return;
                }
            } else {
                write(IOUtil.cbufOfInt[(int) l]);

                return;
            }
        }

        write(String.valueOf(l));
    }

    public void write(float f) throws IOException {
        write(String.valueOf(f));
    }

    public void write(double d) throws IOException {
        write(String.valueOf(d));
    }

    public void write(Date date) throws IOException {
        N.format(date, null, null, this);
    }

    public void write(Calendar c) throws IOException {
        N.format(c, null, null, this);
    }

    public void write(XMLGregorianCalendar c) throws IOException {
        N.format(c, null, null, this);
    }

    @Override
    public void write(String st) throws IOException {
        if (st == null) {
            write(N.NULL_CHAR_ARRAY);
        } else {
            write(st, 0, st.length());
        }
    }

    @Override
    public void write(String st, int off, int len) throws IOException {
        if (st == null) {
            write(N.NULL_CHAR_ARRAY, off, len);

            return;
        }

        //        synchronized (lock) {
        if (len > DEFAULT_BUFFER_SIZE) {
            // Don't permanently allocate very large buffers.
            char[] cbuf = new char[len];
            st.getChars(off, (off + len), cbuf, 0);
            write(cbuf, 0, len);
        } else {
            if (len > (DEFAULT_BUFFER_SIZE - nextChar)) {
                flushBuffer();
            }

            if (cbuff == null) {
                cbuff = createBuffer();
            }

            st.getChars(off, (off + len), cbuff, nextChar);
            nextChar += len;
        }

        //        }
    }

    @Override
    public void write(char[] cbuf, int off, int len) throws IOException {
        //        synchronized (lock) {
        if (len > (DEFAULT_BUFFER_SIZE - nextChar)) {
            if (nextChar > 0) {
                flushBuffer();
            }

            doWrite(cbuf, off, len);
        } else {
            if (cbuff == null) {
                cbuff = createBuffer();
            }

            System.arraycopy(cbuf, off, cbuff, nextChar, len);
            nextChar += len;
        }

        //        }
    }

    protected void doWrite(char[] cbuf, int off, int len)
        throws IOException {
        if (sb == null) {
            out.write(cbuf, off, len);
        } else {
            sb.append(cbuf, off, len);
        }
    }

    /**
     * Writes a line separator.  The line separator string is defined by the
     * system property <tt>line.separator</tt>, and is not necessarily a single
     * newline ('\n') character.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void newLine() throws IOException {
        write(N.LINE_SEPARATOR);
    }

    /**
     * @deprecated
     * Replaced with write(char).
     */
    @Override
    public Writer append(CharSequence csq) throws IOException {
        return super.append(csq);
    }

    /**
     * @deprecated
     * Replaced with write(char).
     */
    @Override
    public Writer append(CharSequence csq, int start, int end)
        throws IOException {
        return super.append(csq, start, end);
    }

    /**
     * @deprecated
     * Replaced with write(char).
     */
    @Override
    public Writer append(char c) throws IOException {
        return super.append(c);
    }

    @Override
    public void flush() throws IOException {
        //        synchronized (lock) {
        flushBuffer();

        recycleBuffer(cbuff);
        cbuff = null;
        nextChar = 0;

        recycleCharArray(cArray);
        cArray = null;

        if (out != null) {
            out.flush();
        }

        //        }
    }

    @Override
    public void close() throws IOException {
        //        synchronized (lock) {
        try {
            if ((sb != null) || (out != null)) {
                flushBuffer();
            }
        } finally {
            recycleBuffer(cbuff);
            cbuff = null;
            nextChar = 0;

            recycleCharArray(cArray);
            cArray = null;

            sb = null;

            IOUtil.close(out);
        }

        //        }
    }

    @Override
    public String toString() {
        if ((sb == null) || (sb.length() == 0)) {
            if (out == null) {
                return (nextChar == 0) ? N.EMPTY_STRING : String.valueOf(cbuff, 0, nextChar);
            } else {
                try {
                    flushBuffer();
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }

                return out.toString();
            }
        } else {
            try {
                flushBuffer();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }

            return sb.toString();
        }
    }

    boolean isStringBuilderRecycable() {
        return isStringBuilderRecycable;
    }

    void setStringBuilderRecycable(boolean isStringBuilderRecycable) {
        this.isStringBuilderRecycable = isStringBuilderRecycable;
    }

    StringBuilder getStringBuilder() {
        return sb;
    }

    void setStringBuilder(StringBuilder sb) {
        this.sb = sb;
        lock = sb;
    }

    void setWriter(OutputStream os) {
        setWriter(new OutputStreamWriter(os));
    }

    void setWriter(Writer writer) {
        this.out = writer;
        lock = writer;
    }

    void _reset() {
        //        synchronized (lock) {
        recycleBuffer(cbuff);
        cbuff = null;
        nextChar = 0;

        recycleCharArray(cArray);
        cArray = null;

        sb = null;
        out = null;

        isStringBuilderRecycable = false;

        //        }
    }

    protected char[] createBuffer() {
        synchronized (bufferPool) {
            return (bufferPool.size() == 0) ? new char[DEFAULT_BUFFER_SIZE] : bufferPool.remove(bufferPool.size() - 1);
        }
    }

    protected void recycleBuffer(char[] cbuff) {
        if ((cbuff != null) && (cbuff.length == DEFAULT_BUFFER_SIZE)) {
            synchronized (bufferPool) {
                bufferPool.add(cbuff);
            }
        }
    }

    protected char[] createCharArray() {
        synchronized (arrayPool) {
            return (arrayPool.size() == 0) ? new char[DEFAULT_ARRAY_SIZE] : arrayPool.remove(arrayPool.size() - 1);
        }
    }

    protected void recycleCharArray(char[] cArray) {
        if ((cArray != null) && (cArray.length == DEFAULT_ARRAY_SIZE)) {
            synchronized (arrayPool) {
                arrayPool.add(cArray);
            }
        }
    }

    protected void flushBuffer() throws IOException {
        //        synchronized (lock) {
        if (nextChar == 0) {
            return;
        }

        doWrite(cbuff, 0, nextChar);

        nextChar = 0;

        //        }
    }

    protected void writeHex(int ch) throws IOException {
        write("&#x");
        write(Integer.toHexString(ch));
        write(';');
    }
}
