package com.google.gson.util;

import java.util.Calendar;
import java.util.HashMap;
import java.util.Map;
public enum CalendarUnit {
    /**
     * it has the same int value as {@code Calendar.MILLISECOND}
     * @see Calendar#MILLISECOND
     */
    MILLISECOND(Calendar.MILLISECOND), 
    /**
     * it has the same int value as {@code Calendar.SECOND}
     * @see Calendar#SECOND
     */
    SECOND(Calendar.SECOND), 
    /**
     * it has the same int value as {@code Calendar.MINUTE}
     * @see Calendar#MINUTE
     */
    MINUTE(Calendar.MINUTE), 
    /**
     * it has the same int value as {@code Calendar.HOUR_OF_DAY}
     * @see Calendar#HOUR
     */
    HOUR(Calendar.HOUR_OF_DAY), 
    /**
     * it has the same int value as {@code Calendar.DAY_OF_MONTH}
     * @see Calendar#DAY_OF_MONTH
     */
    DAY(Calendar.DAY_OF_MONTH), 
    /**
     * it has the same int value as {@code Calendar.WEEK_OF_YEAR}
     * @see Calendar#WEEK_OF_YEAR
     */
    WEEK(Calendar.WEEK_OF_YEAR), 
    /**
     * it has the same int value as {@code Calendar.MONTH}
     * @see Calendar#MONTH
     */
    MONTH(Calendar.MONTH), 
    /**
     * it has the same int value as {@code Calendar.YEAR}
     * @see Calendar#YEAR
     */
    YEAR(Calendar.YEAR);
    private static Map<Integer, CalendarUnit> m = new HashMap<Integer, CalendarUnit>();

    static {
        for (CalendarUnit value : CalendarUnit.values()) {
            m.put(value.intValue, value);
        }
    }

    private int intValue;

    CalendarUnit(int intValue) {
        this.intValue = intValue;
    }

    public static CalendarUnit valueOf(int intValue) {
        CalendarUnit result = m.get(intValue);

        if (result == null) {
            throw new IllegalArgumentException("No defined CalendarUnit mapping to value: " + intValue);
        }

        return result;
    }

    public int intValue() {
        return intValue;
    }

    public long toMillis(int amount) {
        switch (this) {
        case MILLISECOND:
            return amount;

        case SECOND:
            return amount * 1000L;

        case MINUTE:
            return amount * 1000L * 60L;

        case HOUR:
            return amount * 1000L * 60L * 60L;

        case DAY:
            return amount * 1000L * 60L * 60L * 24L;

        case WEEK:
            return amount * 1000L * 60L * 60L * 24L * 7L;

        default:
            throw new RuntimeException("Unsupported unit: " + this);
        }
    }
}
