/*
 * @(#)FileUtil.java        07/04/05
 *
 * Copyright (c) 2007 li haiyang. All rights reserved.
 */
package com.google.gson.util;


import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.lang.management.ManagementFactory;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;


/**
 *
 * @author li haiyang
 *
 * @version $Revision: 0.8 $
 */
public class IOUtil {
    private static final String ZIP = ".zip";
    private static final String GZ = ".gz";

    /**
     * Field BUFFER_SIZE.
     * (value is 10240)
     */
    private static final int BUFFER_SIZE = 1024;
    static final char[] _2147483648_CHAR_ARRAY = "-2147483648".toCharArray();
    static final char[] _9223372036854775808_CHAR_ARRAY = "-9223372036854775808".toCharArray();
    static final int maxCBufIntValue;
    static final char[][] cbufOfInt;

    static {
        int temp = (int) ((ManagementFactory.getMemoryMXBean().getHeapMemoryUsage().getMax() / 1000000000) * 100000);
        maxCBufIntValue = ((temp < -Short.MIN_VALUE) ? (-Short.MIN_VALUE) : temp) + 1;

        int len = IOUtil.maxCBufIntValue + 1;
        cbufOfInt = new char[len][];

        for (int i = 0; i < len; i++) {
            cbufOfInt[i] = String.valueOf(i).toCharArray();
        }
    }

    /**
     * Constructor for FileUtil.
     */
    private IOUtil() {
        // no instance;
    }

    public static String readString(File file) {
        Reader reader = null;

        try {
            reader = new FileReader(file);

            return readString(reader);
        } catch (FileNotFoundException e) {
            throw new RuntimeException(e);
        } finally {
            IOUtil.close(reader);
        }
    }

    public static String readString(InputStream is) {
        Reader reader = null;

        try {
            reader = new InputStreamReader(is);

            return readString(reader);
        } finally {
            // close(reader);
        }
    }

    public static String readString(Reader reader) {
        StringBuilder sb = null;
        char[] cbuf = ObjectFactory.createCharArrayBuffer();
        int len = cbuf.length;

        try {
            int count = reader.read(cbuf);

            if ((count == -1) || (count == 0)) {
                return N.EMPTY_STRING;
            } else {
                if (count < len) {
                    return String.valueOf(cbuf, 0, count);
                } else {
                    sb = ObjectFactory.createStringBuilder(ObjectFactory.POOLABLE_SIZE * 2);
                    sb.append(cbuf);

                    do {
                        count = reader.read(cbuf);

                        if (count <= 0) {
                            break;
                        }

                        sb.append(cbuf, 0, count);
                    } while (count >= len);

                    return sb.toString();
                }
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            ObjectFactory.recycle(cbuf);
            ObjectFactory.recycle(sb);
        }
    }

    public static char[] readChars(File file) {
        Reader reader = null;

        try {
            reader = new FileReader(file);

            return readChars(reader);
        } catch (FileNotFoundException e) {
            throw new RuntimeException(e);
        } finally {
            IOUtil.close(reader);
        }
    }

    public static char[] readChars(InputStream is) {
        Reader reader = null;

        try {
            reader = new InputStreamReader(is);

            return readChars(reader);
        } finally {
            // close(reader);
        }
    }

    public static char[] readChars(Reader reader) {
        StringBuilder sb = null;
        char[] cbuf = ObjectFactory.createCharArrayBuffer();
        int len = cbuf.length;

        try {
            int count = reader.read(cbuf);

            if ((count == -1) || (count == 0)) {
                return new char[0];
            } else {
                if (count < len) {
                    return Arrays.copyOf(cbuf, count);
                } else {
                    sb = ObjectFactory.createStringBuilder(ObjectFactory.POOLABLE_SIZE * 2);
                    sb.append(cbuf);

                    do {
                        count = reader.read(cbuf);

                        if (count <= 0) {
                            break;
                        }

                        sb.append(cbuf, 0, count);
                    } while (count >= len);

                    char[] chars = new char[sb.length()];

                    sb.getChars(0, sb.length(), chars, 0);

                    return chars;
                }
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            ObjectFactory.recycle(cbuf);
            ObjectFactory.recycle(sb);
        }
    }

    public static List<String> readLines(File file) {
        InputStream is = createInputStream(file);

        try {
            return readLines(is);
        } finally {
            close(is);
        }
    }

    public static List<String> readLines(InputStream is) {
        return readLines(new InputStreamReader(is));
    }

    public static List<String> readLines(Reader reader) {
        List<String> lines = N.asList();
        BufferedReader br = ObjectFactory.createBufferedReader(reader);

        try {
            String line = br.readLine();

            while (line != null) {
                lines.add(line);
                line = br.readLine();
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        } finally {
            ObjectFactory.recycle(br);
        }

        return lines;
    }

    public static byte[] readBytes(File file) {
        byte[] bytes = new byte[(int) file.length()];
        InputStream is = null;

        try {
            is = new FileInputStream(file);

            is.read(bytes);
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            close(is);
        }

        return bytes;
    }

    public static byte[] readBytes(InputStream is) {
        ByteArrayOutputStream os = null;
        byte[] bytes = ObjectFactory.createByteArrayBuffer();
        int len = bytes.length;

        try {
            int count = is.read(bytes);

            if (count == -1) {
                return new byte[0];
            } else if (count < len) {
                return Arrays.copyOf(bytes, count);
            } else {
                os = ObjectFactory.createByteArrayOutputStream(ObjectFactory.POOLABLE_SIZE * 2);
                os.write(bytes);

                do {
                    count = is.read(bytes);

                    if (count <= 0) {
                        break;
                    }

                    os.write(bytes, 0, count);
                } while (count >= len);

                return os.toByteArray();
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            ObjectFactory.recycle(bytes);
            ObjectFactory.recycle(os);
        }
    }

    public static void writeInt(Writer writer, int i) throws IOException {
        if (i < IOUtil.maxCBufIntValue) {
            if (i < 0) {
                if (i == Integer.MIN_VALUE) {
                    writer.write(IOUtil._2147483648_CHAR_ARRAY);

                    return;
                } else if (-i < IOUtil.maxCBufIntValue) {
                    writer.write('-');
                    writer.write(cbufOfInt[-i]);

                    return;
                }
            } else {
                writer.write(cbufOfInt[i]);

                return;
            }
        }

        writer.write(String.valueOf(i));
    }

    public static void writeLong(Writer writer, long l)
        throws IOException {
        if (l < IOUtil.maxCBufIntValue) {
            if (l < 0) {
                if (l == Long.MIN_VALUE) {
                    writer.write(IOUtil._9223372036854775808_CHAR_ARRAY);

                    return;
                } else if (-l < IOUtil.maxCBufIntValue) {
                    writer.write('-');
                    writer.write(cbufOfInt[(int) -l]);

                    return;
                }
            } else {
                writer.write(cbufOfInt[(int) l]);

                return;
            }
        }

        writer.write(String.valueOf(l));
    }

    public static void writeString(String st, File out) {
        writeChars(st.toCharArray(), out);
    }

    public static void writeString(String st, OutputStream out) {
        writeChars(st.toCharArray(), out);
    }

    public static void writeString(String st, Writer out) {
        writeChars(st.toCharArray(), out);
    }

    public static void writeChars(char[] chars, File out) {
        Writer writer = null;

        try {
            if (out.exists()) {
                out.createNewFile();
            }

            writer = new FileWriter(out);
            writeChars(chars, writer);
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            close(writer);
        }
    }

    public static void writeChars(char[] chars, OutputStream out) {
        Writer writer = new OutputStreamWriter(out);

        try {
            writeChars(chars, writer);
        } finally {
            // close(writer);
        }
    }

    public static void writeChars(char[] chars, Writer out) {
        try {
            out.write(chars);
            out.flush();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public static void writeBytes(byte[] bytes, File out) {
        OutputStream os = null;

        try {
            if (out.exists()) {
                out.createNewFile();
            }

            writeBytes(bytes, os);
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            close(os);
        }
    }

    public static void writeBytes(byte[] bytes, OutputStream out) {
        try {
            out.write(bytes);
            out.flush();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public static InputStream createInputStream(File file) {
        InputStream is = null;

        try {
            is = new FileInputStream(file);

            if (file.getName().endsWith(GZ)) {
                is = new GZIPInputStream(is);
            } else if (file.getName().endsWith(ZIP)) {
                ZipFile zf = new ZipFile(file);
                ZipEntry ze = (ZipEntry) zf.entries().nextElement();
                is = zf.getInputStream(ze);
            }
        } catch (IOException e) {
            close(is);
            throw new RuntimeException(e);
        }

        return is;
    }

    public static void close(InputStream is) {
        if (is != null) {
            try {
                is.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    public static void close(OutputStream os) {
        if (os != null) {
            try {
                os.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    public static void close(Reader reader) {
        if (reader != null) {
            try {
                reader.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    public static void close(Writer write) {
        if (write != null) {
            try {
                write.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    public static void close(ZipFile zip) {
        if (zip != null) {
            try {
                zip.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    public static void zip(File sourceFile, File targetFile, boolean deleteTargetFileIfExists) {
        if (targetFile.exists() && deleteTargetFileIfExists) {
            targetFile.delete();
        }

        ZipOutputStream zos = null;

        try {
            zos = new ZipOutputStream(new FileOutputStream(targetFile));
            zipDir(sourceFile, zos, targetFile);
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            close(zos);
        }
    }

    public static void unzip(File sourceZipFile, File targetFile) {
        ZipFile zip = null;
        ZipEntry ze = null;
        OutputStream os = null;
        InputStream is = null;

        try {
            zip = new ZipFile(sourceZipFile);

            byte[] buf = new byte[BUFFER_SIZE];
            Enumeration<? extends ZipEntry> entryEnum = zip.entries();

            while (entryEnum.hasMoreElements()) {
                ze = (ZipEntry) entryEnum.nextElement();

                if (ze.isDirectory()) {
                    continue;
                }

                os = new BufferedOutputStream(new FileOutputStream(getAbsolutePath(targetFile, ze.getName())));
                is = new BufferedInputStream(zip.getInputStream(ze));

                int readLen = is.read(buf, 0, BUFFER_SIZE);

                while (readLen != -1) {
                    os.write(buf, 0, readLen);
                    readLen = is.read(buf, 0, BUFFER_SIZE);
                }

                close(is);
                is = null;
                close(os);
                os = null;
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            close(zip);
            close(is);
            close(os);
        }
    }

    public static void deleteDir(File file) {
        if (file.isDirectory()) {
            String[] children = file.list();

            for (int i = 0; i < children.length; i++) {
                deleteDir(new File(file, children[i]));
            }
        }

        file.delete();
    }

    public static void splitFile(File file, int lineNum) {
        InputStream is = createInputStream(file);
        BufferedReader br = ObjectFactory.createBufferedReader(is);
        BufferedWriter bw = null;
        int lineCounter = 0;
        String line = null;

        try {
            line = br.readLine();

            while (line != null) {
                if ((lineCounter % lineNum) == 0) {
                    if (bw != null) {
                        bw.flush();
                        bw.close();                        
                        ObjectFactory.recycle(bw);
                        bw = null;
                    }

                    bw = ObjectFactory.createBufferedWriter(new FileWriter(
                                new File(file.getAbsolutePath() + "_" + (lineCounter / lineNum) + ".txt")));
                }

                bw.write(line);
                bw.write(Character.LINE_SEPARATOR);
                lineCounter++;

                line = br.readLine();
            }

            if (bw != null) {
                bw.flush();
                bw.close();
                ObjectFactory.recycle(bw);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            close(is);
            ObjectFactory.recycle(br);
        }
    }

    private static void zipDir(File sourceFile, ZipOutputStream zos, File zipFile)
        throws IOException, FileNotFoundException {
        if (sourceFile.isFile()) {
            zipFile(sourceFile, null, zos, zipFile);
        } else {
            List<File> subFileList = getSubFiles(sourceFile, true);

            // subFileList.add(sourceFile);
            for (int i = 0; i < subFileList.size(); i++) {
                File subFile = (File) subFileList.get(i);
                zipFile(subFile, sourceFile, zos, zipFile);
            }
        }
    }

    private static void zipFile(File file, File sourceDir, ZipOutputStream zos, File zipFile)
        throws IOException, FileNotFoundException {
        if (file.equals(zipFile)) {
            return;
        }

        byte[] buf = new byte[BUFFER_SIZE];
        ZipEntry ze = null;
        String relativeFileName = null;

        if (sourceDir == null) {
            relativeFileName = file.getName();
        } else {
            relativeFileName = getRelativePath(sourceDir, file);
        }

        ze = new ZipEntry(relativeFileName);
        ze.setSize(file.length());
        ze.setTime(file.lastModified());
        zos.putNextEntry(ze);

        InputStream is = new BufferedInputStream(new FileInputStream(file));

        try {
            int readLen = is.read(buf, 0, BUFFER_SIZE);

            while (readLen != -1) {
                zos.write(buf, 0, readLen);
                readLen = is.read(buf, 0, BUFFER_SIZE);
            }
        } finally {
            close(is);
        }
    }

    private static List<File> getSubFiles(File baseDir, boolean isRecersive) {
        List<File> result = new ArrayList<File>();

        if (baseDir.isFile()) {
            result.add(baseDir);
        } else {
            File[] subFiles = baseDir.listFiles();

            for (int i = 0; i < subFiles.length; i++) {
                if (subFiles[i].isFile()) {
                    result.add(subFiles[i]);
                }

                if (subFiles[i].isDirectory()) {
                    result.addAll(getSubFiles(subFiles[i], isRecersive));
                }
            }
        }

        return result;
    }

    private static String getAbsolutePath(File parentDir, String relativeFilePath)
        throws IOException {
        String newRelativePath = "";

        for (int i = 0; i < relativeFilePath.length(); i++) {
            char c = relativeFilePath.charAt(i);

            if ((c == '\\') || (c == '/')) {
                newRelativePath += File.separator;
            } else {
                newRelativePath += c;
            }
        }

        relativeFilePath = newRelativePath;

        String path = parentDir.getAbsolutePath() + File.separator + relativeFilePath;

        File dir = new File(path.substring(0, path.lastIndexOf(File.separator)));

        if (!dir.exists()) {
            dir.mkdirs();
        }

        return path;
    }

    private static String getRelativePath(File parentDir, File file) {
        if (file.equals(parentDir)) {
            return file.getName();
        } else {
            return file.getAbsolutePath().substring(parentDir.getAbsolutePath().length() + 1);
        }
    }
}
