package com.google.gson.util;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Proxy;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.charset.Charset;
import java.rmi.server.UID;
import java.security.SecureRandom;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.text.DecimalFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Deque;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Queue;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TimeZone;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.DelayQueue;
import java.util.concurrent.Delayed;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.FutureTask;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;

import org.w3c.dom.NodeList;

import com.sun.org.apache.xerces.internal.impl.dv.util.Base64;


@SuppressWarnings("unchecked")
public final class N {
    static final Logger logger = Logger.getLogger(N.class.getCanonicalName());
    public static final String HOST_NAME;

    static {
        String temp = "unknown";

        try {
            InetAddress addr = InetAddress.getLocalHost();
            temp = addr.getHostName();
        } catch (UnknownHostException e) {
            logger.warning("Failed to get host name due to error: " + e.getMessage());
        }

        HOST_NAME = temp;
    }

    /**
     * Installation directory for Java Runtime Environment (JRE)
     */
    public static final String JAVA_HOME = System.getProperty("java.home");

    /**
     * JRE version number
     */
    public static final String JAVA_VERSION = System.getProperty("java.version");

    /**
     * User working directory
     */
    public static final String USER_DIR = System.getProperty("user.dir");

    /**
     * Path separator character used in java.class.path
     */
    public static final String PATH_SEPARATOR = System.getProperty("path.separator");

    /**
     * FILE_SEPARATOR
     */
    public static final String FILE_SEPARATOR = File.separator; // System.getProperty("file.separator");

    /**
     * Sequence used by operating system to separate lines in text files
     */
    public static final String LINE_SEPARATOR = System.getProperty("line.separator");

    // ...
    public static final Charset UTF_8 = Charset.forName("UTF-8");
    public static final Charset ASCII = Charset.forName("US-ASCII");
    public static final Charset ISO_8859_1 = Charset.forName("ISO-8859-1");
    public static final String NULL_STRING = "null";
    public static final char[] NULL_CHAR_ARRAY = NULL_STRING.toCharArray();
    public static final String EMPTY_STRING = "";

    /**
    * Field TRUE. (value is ""true"")
    */
    public static final String TRUE = Boolean.TRUE.toString();

    /**
     * Field FALSE. (value is ""false"")
     */
    public static final String FALSE = Boolean.FALSE.toString();

    // ...
    public static final char[] TRUE_CHAR_ARRAY = Boolean.TRUE.toString().toCharArray();
    public static final char[] FALSE_CHAR_ARRAY = Boolean.FALSE.toString().toCharArray();

    /**
    * Field GET
    */
    public static final String GET = "get";

    /**
     * Field SET
     */
    public static final String SET = "set";

    /**
     * Field IS
     */
    public static final String IS = "is";

    /**
     * Field HAS
     */
    public static final String HAS = "has";

    //... it has to be big enougth to make it's safety to add element to ArrayBlockingQueue. 
    private static final int SMALL_POOL_SIZE = 1000;
    private static final int BIG_POOL_SIZE = 3000;

    // ...
    private static final SecureRandom numberGenerator = new SecureRandom();

    /**
     * Date format.
     */
    public static final String DEFAULT_DATETIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss'Z'";
    public static final String LOCAL_DATETIME_FORMAT = "yyyy-MM-dd HH:mm:ss";
    public static final String DEFAULT_TIMESTAMP_FORMAT = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";
    private static final Map<String, Queue<SimpleDateFormat>> SDF_POOL = N.asConcurrentHashMap();
    private static final Map<TimeZone, Queue<Calendar>> calendarPool = N.asConcurrentHashMap();
    private static final Queue<Calendar> utcCalendarPool = N.newArrayBlockingQueue(SMALL_POOL_SIZE);
    private static final TimeZone UTC_TIME_ZONE = TimeZone.getTimeZone("UTC");
    private static final TimeZone LOCAL_TIME_ZONE;

    static {
        LOCAL_TIME_ZONE = Calendar.getInstance().getTimeZone();
    }

    private static final DatatypeFactory dataTypeFactory;

    static {
        DatatypeFactory temp = null;

        try {
            temp = DatatypeFactory.newInstance();
        } catch (DatatypeConfigurationException e) {
            logger.warning("Failed to initialize XMLGregorianCalendarType: " + e.getMessage());
        }

        dataTypeFactory = temp;
    }

    // ...
    private static final char[][][] cbufOfSTDInt = new char[5][][];

    static {
        for (int i = 0, j = 1; i < 5; i++, j = j * 10) {
            cbufOfSTDInt[i] = new char[j][];

            for (int k = 0; k < j; k++) {
                if (i == 1) {
                    cbufOfSTDInt[i][k] = String.valueOf(k).toCharArray();
                } else if (i == 2) {
                    if (k < 10) {
                        cbufOfSTDInt[i][k] = ("0" + String.valueOf(k)).toCharArray();
                    } else {
                        cbufOfSTDInt[i][k] = String.valueOf(k).toCharArray();
                    }
                } else if (i == 3) {
                    if (k < 10) {
                        cbufOfSTDInt[i][k] = ("00" + String.valueOf(k)).toCharArray();
                    } else if (k < 100) {
                        cbufOfSTDInt[i][k] = ("0" + String.valueOf(k)).toCharArray();
                    } else {
                        cbufOfSTDInt[i][k] = String.valueOf(k).toCharArray();
                    }
                } else if (i == 4) {
                    if (k < 10) {
                        cbufOfSTDInt[i][k] = ("000" + String.valueOf(k)).toCharArray();
                    } else if (k < 100) {
                        cbufOfSTDInt[i][k] = ("00" + String.valueOf(k)).toCharArray();
                    } else if (k < 1000) {
                        cbufOfSTDInt[i][k] = ("0" + String.valueOf(k)).toCharArray();
                    } else {
                        cbufOfSTDInt[i][k] = String.valueOf(k).toCharArray();
                    }
                }
            }
        }
    }

    //...
    private static final Set<Class<?>> PRIMARY_TYPE = N.asSet();

    static {
        PRIMARY_TYPE.add(boolean.class);
        PRIMARY_TYPE.add(char.class);
        PRIMARY_TYPE.add(byte.class);
        PRIMARY_TYPE.add(short.class);
        PRIMARY_TYPE.add(int.class);
        PRIMARY_TYPE.add(long.class);
        PRIMARY_TYPE.add(float.class);
        PRIMARY_TYPE.add(double.class);
        PRIMARY_TYPE.add(Boolean.class);
        PRIMARY_TYPE.add(Character.class);
        PRIMARY_TYPE.add(Byte.class);
        PRIMARY_TYPE.add(Short.class);
        PRIMARY_TYPE.add(Integer.class);
        PRIMARY_TYPE.add(Long.class);
        PRIMARY_TYPE.add(Float.class);
        PRIMARY_TYPE.add(Double.class);
    }

    private static final Map<Class<?>, Object> DEFAULT_VALUE_FOR_PRIMARY_TYPE = N.asMap();

    static {
        DEFAULT_VALUE_FOR_PRIMARY_TYPE.put(boolean.class, false);
        DEFAULT_VALUE_FOR_PRIMARY_TYPE.put(char.class, (char) 0);
        DEFAULT_VALUE_FOR_PRIMARY_TYPE.put(byte.class, (byte) 0);
        DEFAULT_VALUE_FOR_PRIMARY_TYPE.put(short.class, (short) 0);
        DEFAULT_VALUE_FOR_PRIMARY_TYPE.put(int.class, (int) 0);
        DEFAULT_VALUE_FOR_PRIMARY_TYPE.put(long.class, (long) 0L);
        DEFAULT_VALUE_FOR_PRIMARY_TYPE.put(float.class, (float) 0f);
        DEFAULT_VALUE_FOR_PRIMARY_TYPE.put(double.class, (double) 0d);
    }

    private static final Map<Class<?>, Integer> CLASS_ENUM_ORDER = N.asMap();

    static {
        CLASS_ENUM_ORDER.put(boolean.class, 1);
        CLASS_ENUM_ORDER.put(char.class, 2);
        CLASS_ENUM_ORDER.put(byte.class, 3);
        CLASS_ENUM_ORDER.put(short.class, 4);
        CLASS_ENUM_ORDER.put(int.class, 5);
        CLASS_ENUM_ORDER.put(long.class, 6);
        CLASS_ENUM_ORDER.put(float.class, 7);
        CLASS_ENUM_ORDER.put(double.class, 8);

        CLASS_ENUM_ORDER.put(Boolean.class, 9);
        CLASS_ENUM_ORDER.put(Character.class, 10);
        CLASS_ENUM_ORDER.put(Byte.class, 11);
        CLASS_ENUM_ORDER.put(Short.class, 12);
        CLASS_ENUM_ORDER.put(Integer.class, 13);
        CLASS_ENUM_ORDER.put(Long.class, 14);
        CLASS_ENUM_ORDER.put(Float.class, 15);
        CLASS_ENUM_ORDER.put(Double.class, 16);

        CLASS_ENUM_ORDER.put(BigInteger.class, 17);
        CLASS_ENUM_ORDER.put(BigDecimal.class, 18);

        CLASS_ENUM_ORDER.put(java.util.Date.class, 19);
        CLASS_ENUM_ORDER.put(Date.class, 20);
        CLASS_ENUM_ORDER.put(Time.class, 21);
        CLASS_ENUM_ORDER.put(Timestamp.class, 22);

        CLASS_ENUM_ORDER.put(Calendar.class, 23);
        CLASS_ENUM_ORDER.put(GregorianCalendar.class, 24);
        CLASS_ENUM_ORDER.put(XMLGregorianCalendar.class, 25);
    }

    static final Set<Class<?>> abstractPropClass = N.asSet();

    static {
        abstractPropClass.add(Collection.class);
        abstractPropClass.add(List.class);
        abstractPropClass.add(Set.class);
        abstractPropClass.add(Queue.class);
        abstractPropClass.add(Map.class);
    }

    /**
     * Field registeredJAXBClassList.
     */
    private static final Map<Class<?>, Boolean> registeredXMLBindingClassList = N.newConcurrentHashMap(SMALL_POOL_SIZE);

    /**
     * Field entityDeclaredPropGetMethodList.
     */
    private static final Map<Class<?>, Map<String, Method>> entityDeclaredPropGetMethodList = N.newConcurrentHashMap(SMALL_POOL_SIZE);

    /**
     * Field entityDeclaredPropSetMethodList.
     */
    private static final Map<Class<?>, Map<String, Method>> entityDeclaredPropSetMethodList = N.newConcurrentHashMap(SMALL_POOL_SIZE);

    /**
     * Field entityPropGetMethodPool.
     */
    private static final Map<Class<?>, Map<String, Method>> entityPropGetMethodPool = N.newConcurrentHashMap(SMALL_POOL_SIZE);

    /**
     * Field entityPropSetMethodPool.
     */
    private static final Map<Class<?>, Map<String, Method>> entityPropSetMethodPool = N.newConcurrentHashMap(SMALL_POOL_SIZE);

    /**
     * Field propMethodName.
     */
    private static final Map<Method, String> methodPropNamePool = N.newConcurrentHashMap(BIG_POOL_SIZE);
    private static final Map<Method, Class<?>[]> methodTypeArgumentsPool = N.newConcurrentHashMap(BIG_POOL_SIZE);

    /**
     * Field DUMMY object.
     */
    public static final Method DUMMY_METHOD = N.getDeclaredMethod(N.class, "dummyMethod");
    public static final Class<?> DUMMY_CLASS = DummyClass.class;
    public static final Class<?>[] DUMMY_CLASSES = new Class<?>[] { DUMMY_CLASS };

    // reserved words.
    private static final Map<String, String> keyWords = N.newHashMap(100);

    static {
        keyWords.put("class", "clazz");
    }

    // formalized property name list.
    private static final Map<String, String> formalizedPropNamePool = N.newConcurrentHashMap(BIG_POOL_SIZE);
    private static final Map<String, Class<?>> BUILT_IN_TYPE = N.newConcurrentHashMap(SMALL_POOL_SIZE);

    static {
        BUILT_IN_TYPE.put(boolean.class.getCanonicalName(), boolean.class);
        BUILT_IN_TYPE.put(char.class.getCanonicalName(), char.class);
        BUILT_IN_TYPE.put(byte.class.getCanonicalName(), byte.class);
        BUILT_IN_TYPE.put(short.class.getCanonicalName(), short.class);
        BUILT_IN_TYPE.put(int.class.getCanonicalName(), int.class);
        BUILT_IN_TYPE.put(long.class.getCanonicalName(), long.class);
        BUILT_IN_TYPE.put(float.class.getCanonicalName(), float.class);
        BUILT_IN_TYPE.put(double.class.getCanonicalName(), double.class);

        BUILT_IN_TYPE.put(boolean.class.getSimpleName(), boolean.class);
        BUILT_IN_TYPE.put(char.class.getSimpleName(), char.class);
        BUILT_IN_TYPE.put(byte.class.getSimpleName(), byte.class);
        BUILT_IN_TYPE.put(short.class.getSimpleName(), short.class);
        BUILT_IN_TYPE.put(int.class.getSimpleName(), int.class);
        BUILT_IN_TYPE.put(long.class.getSimpleName(), long.class);
        BUILT_IN_TYPE.put(float.class.getSimpleName(), float.class);
        BUILT_IN_TYPE.put(double.class.getSimpleName(), double.class);

        BUILT_IN_TYPE.put(Boolean.class.getCanonicalName(), Boolean.class);
        BUILT_IN_TYPE.put(Character.class.getCanonicalName(), Character.class);
        BUILT_IN_TYPE.put(Byte.class.getCanonicalName(), Byte.class);
        BUILT_IN_TYPE.put(Short.class.getCanonicalName(), Short.class);
        BUILT_IN_TYPE.put(Integer.class.getCanonicalName(), Integer.class);
        BUILT_IN_TYPE.put(Long.class.getCanonicalName(), Long.class);
        BUILT_IN_TYPE.put(Float.class.getCanonicalName(), Float.class);
        BUILT_IN_TYPE.put(Double.class.getCanonicalName(), Double.class);

        BUILT_IN_TYPE.put(Boolean.class.getSimpleName(), Boolean.class);
        BUILT_IN_TYPE.put(Character.class.getSimpleName(), Character.class);
        BUILT_IN_TYPE.put(Byte.class.getSimpleName(), Byte.class);
        BUILT_IN_TYPE.put(Short.class.getSimpleName(), Short.class);
        BUILT_IN_TYPE.put(Integer.class.getSimpleName(), Integer.class);
        BUILT_IN_TYPE.put(Long.class.getSimpleName(), Long.class);
        BUILT_IN_TYPE.put(Float.class.getSimpleName(), Float.class);
        BUILT_IN_TYPE.put(Double.class.getSimpleName(), Double.class);

        BUILT_IN_TYPE.put(String.class.getCanonicalName(), String.class);
        BUILT_IN_TYPE.put(Object.class.getCanonicalName(), Object.class);
        BUILT_IN_TYPE.put(Class.class.getCanonicalName(), Class.class);
        BUILT_IN_TYPE.put(Enum.class.getCanonicalName(), Enum.class);
        BUILT_IN_TYPE.put(String.class.getSimpleName(), String.class);
        BUILT_IN_TYPE.put(Object.class.getSimpleName(), Object.class);
        BUILT_IN_TYPE.put(Class.class.getSimpleName(), Class.class);
        BUILT_IN_TYPE.put(Enum.class.getSimpleName(), Enum.class);

        BUILT_IN_TYPE.put(java.util.Date.class.getCanonicalName(), java.util.Date.class);
        BUILT_IN_TYPE.put(Calendar.class.getCanonicalName(), Calendar.class);
        BUILT_IN_TYPE.put(GregorianCalendar.class.getCanonicalName(), GregorianCalendar.class);
        BUILT_IN_TYPE.put(Collection.class.getCanonicalName(), Collection.class);
        BUILT_IN_TYPE.put(List.class.getCanonicalName(), List.class);
        BUILT_IN_TYPE.put(ArrayList.class.getCanonicalName(), ArrayList.class);
        BUILT_IN_TYPE.put(LinkedList.class.getCanonicalName(), LinkedList.class);
        BUILT_IN_TYPE.put(Set.class.getCanonicalName(), Set.class);
        BUILT_IN_TYPE.put(HashSet.class.getCanonicalName(), HashSet.class);
        BUILT_IN_TYPE.put(Map.class.getCanonicalName(), Map.class);
        BUILT_IN_TYPE.put(HashMap.class.getCanonicalName(), HashMap.class);

        BUILT_IN_TYPE.put(Date.class.getCanonicalName(), Date.class);
        BUILT_IN_TYPE.put(Time.class.getCanonicalName(), Time.class);
        BUILT_IN_TYPE.put(Timestamp.class.getCanonicalName(), Timestamp.class);

        BUILT_IN_TYPE.put(boolean[].class.getCanonicalName(), boolean[].class);
        BUILT_IN_TYPE.put(char[].class.getCanonicalName(), char[].class);
        BUILT_IN_TYPE.put(byte[].class.getCanonicalName(), byte[].class);
        BUILT_IN_TYPE.put(short[].class.getCanonicalName(), short[].class);
        BUILT_IN_TYPE.put(int[].class.getCanonicalName(), int[].class);
        BUILT_IN_TYPE.put(long[].class.getCanonicalName(), long[].class);
        BUILT_IN_TYPE.put(float[].class.getCanonicalName(), float[].class);
        BUILT_IN_TYPE.put(double[].class.getCanonicalName(), double[].class);

        BUILT_IN_TYPE.put(Boolean[].class.getCanonicalName(), Boolean[].class);
        BUILT_IN_TYPE.put(Character[].class.getCanonicalName(), Character[].class);
        BUILT_IN_TYPE.put(Byte[].class.getCanonicalName(), Byte[].class);
        BUILT_IN_TYPE.put(Short[].class.getCanonicalName(), Short[].class);
        BUILT_IN_TYPE.put(Integer[].class.getCanonicalName(), Integer[].class);
        BUILT_IN_TYPE.put(Long[].class.getCanonicalName(), Long[].class);
        BUILT_IN_TYPE.put(Float[].class.getCanonicalName(), Float[].class);
        BUILT_IN_TYPE.put(Double[].class.getCanonicalName(), Double[].class);

        BUILT_IN_TYPE.put(String[].class.getCanonicalName(), String[].class);
        BUILT_IN_TYPE.put(Object[].class.getCanonicalName(), Object[].class);
        BUILT_IN_TYPE.put(Class[].class.getCanonicalName(), Class[].class);
        BUILT_IN_TYPE.put(Enum[].class.getCanonicalName(), Enum[].class);

        BUILT_IN_TYPE.put(java.util.Date[].class.getCanonicalName(), java.util.Date[].class);
        BUILT_IN_TYPE.put(Calendar[].class.getCanonicalName(), Calendar[].class);
        BUILT_IN_TYPE.put(GregorianCalendar[].class.getCanonicalName(), GregorianCalendar[].class);
        BUILT_IN_TYPE.put(Collection[].class.getCanonicalName(), Collection[].class);
        BUILT_IN_TYPE.put(List[].class.getCanonicalName(), List[].class);
        BUILT_IN_TYPE.put(ArrayList[].class.getCanonicalName(), ArrayList[].class);
        BUILT_IN_TYPE.put(LinkedList[].class.getCanonicalName(), LinkedList[].class);
        BUILT_IN_TYPE.put(Set[].class.getCanonicalName(), Set[].class);
        BUILT_IN_TYPE.put(HashSet[].class.getCanonicalName(), HashSet[].class);
        BUILT_IN_TYPE.put(Map[].class.getCanonicalName(), Map[].class);
        BUILT_IN_TYPE.put(HashMap[].class.getCanonicalName(), HashMap[].class);

        BUILT_IN_TYPE.put(Date[].class.getCanonicalName(), Date[].class);
        BUILT_IN_TYPE.put(Time[].class.getCanonicalName(), Time[].class);
        BUILT_IN_TYPE.put(Timestamp[].class.getCanonicalName(), Timestamp[].class);

        BUILT_IN_TYPE.put(boolean[][].class.getCanonicalName(), boolean[][].class);
        BUILT_IN_TYPE.put(char[][].class.getCanonicalName(), char[][].class);
        BUILT_IN_TYPE.put(byte[][].class.getCanonicalName(), byte[][].class);
        BUILT_IN_TYPE.put(short[][].class.getCanonicalName(), short[][].class);
        BUILT_IN_TYPE.put(int[][].class.getCanonicalName(), int[][].class);
        BUILT_IN_TYPE.put(long[][].class.getCanonicalName(), long[][].class);
        BUILT_IN_TYPE.put(float[][].class.getCanonicalName(), float[][].class);
        BUILT_IN_TYPE.put(double[][].class.getCanonicalName(), double[][].class);

        BUILT_IN_TYPE.put(Boolean[][].class.getCanonicalName(), Boolean[][].class);
        BUILT_IN_TYPE.put(Character[][].class.getCanonicalName(), Character[][].class);
        BUILT_IN_TYPE.put(Byte[][].class.getCanonicalName(), Byte[][].class);
        BUILT_IN_TYPE.put(Short[][].class.getCanonicalName(), Short[][].class);
        BUILT_IN_TYPE.put(Integer[][].class.getCanonicalName(), Integer[][].class);
        BUILT_IN_TYPE.put(Long[][].class.getCanonicalName(), Long[][].class);
        BUILT_IN_TYPE.put(Float[][].class.getCanonicalName(), Float[][].class);
        BUILT_IN_TYPE.put(Double[][].class.getCanonicalName(), Double[][].class);

        BUILT_IN_TYPE.put(String[][].class.getCanonicalName(), String[][].class);
        BUILT_IN_TYPE.put(Object[][].class.getCanonicalName(), Object[][].class);
        BUILT_IN_TYPE.put(Class[][].class.getCanonicalName(), Class[][].class);
        BUILT_IN_TYPE.put(Enum[][].class.getCanonicalName(), Enum[][].class);

        BUILT_IN_TYPE.put(java.util.Date[][].class.getCanonicalName(), java.util.Date[][].class);
        BUILT_IN_TYPE.put(Calendar[][].class.getCanonicalName(), Calendar[][].class);
        BUILT_IN_TYPE.put(GregorianCalendar[][].class.getCanonicalName(), GregorianCalendar[][].class);
        BUILT_IN_TYPE.put(Collection[][].class.getCanonicalName(), Collection[][].class);
        BUILT_IN_TYPE.put(List[][].class.getCanonicalName(), List[][].class);
        BUILT_IN_TYPE.put(ArrayList[][].class.getCanonicalName(), ArrayList[][].class);
        BUILT_IN_TYPE.put(LinkedList[][].class.getCanonicalName(), LinkedList[][].class);
        BUILT_IN_TYPE.put(Set[][].class.getCanonicalName(), Set[][].class);
        BUILT_IN_TYPE.put(HashSet[][].class.getCanonicalName(), HashSet[][].class);
        BUILT_IN_TYPE.put(Map[][].class.getCanonicalName(), Map[][].class);
        BUILT_IN_TYPE.put(HashMap[][].class.getCanonicalName(), HashMap[][].class);

        BUILT_IN_TYPE.put(Date[][].class.getCanonicalName(), Date[][].class);
        BUILT_IN_TYPE.put(Time[][].class.getCanonicalName(), Time[][].class);
        BUILT_IN_TYPE.put(Timestamp[][].class.getCanonicalName(), Timestamp[][].class);
    }

    private static Map<String, String> SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME = N.asMap();

    static {
        SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.put(boolean.class.getName(), "Z");
        SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.put(char.class.getName(), "C");
        SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.put(byte.class.getName(), "B");
        SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.put(short.class.getName(), "S");
        SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.put(int.class.getName(), "I");
        SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.put(long.class.getName(), "J");
        SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.put(float.class.getName(), "F");
        SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.put(double.class.getName(), "D");
    }

    private static final Map<Class<?>, ClazzParserType> clsParserTypePool = newConcurrentHashMap(BIG_POOL_SIZE);
    private static final Map<String, Class<?>> clsNamePool = N.newConcurrentHashMap(BIG_POOL_SIZE);
    private static final Map<Class<?>, String> simpleNameClassPool = N.newConcurrentHashMap(BIG_POOL_SIZE);
    private static final Map<Class<?>, String> nameClassPool = N.newConcurrentHashMap(BIG_POOL_SIZE);
    private static final Map<Class<?>, String> canonicalNameClassPool = N.newConcurrentHashMap(BIG_POOL_SIZE);
    private static final Map<Class<?>, Class<?>> enclosingClassPool = N.newConcurrentHashMap(BIG_POOL_SIZE);

    // 
    private static final ExecutorService executorService = Executors.newFixedThreadPool(64);

    static {
        Runtime.getRuntime().addShutdownHook(new Thread() {
                public void run() {
                    executorService.shutdown();

                    try {
                        executorService.awaitTermination(180, TimeUnit.SECONDS);
                    } catch (InterruptedException e) {
                        logger.warning("failed to commit task in the queue in N.class due to error: " + e.getMessage());
                    }
                }
            });
    }

    //
    //    private static Class<?>[] getTypeArguments(Class<?> cls) {
    //        java.lang.reflect.Type[] typeArgs = null;
    //        java.lang.reflect.Type[] genericInterfaces = cls.getGenericInterfaces();
    //
    //        if (N.notNullOrEmpty(genericInterfaces)) {
    //            for (java.lang.reflect.Type type : genericInterfaces) {
    //                typeArgs = ((ParameterizedType) type).getActualTypeArguments();
    //
    //                if (N.notNullOrEmpty(typeArgs)) {
    //                    break;
    //                }
    //            }
    //        } else {
    //            java.lang.reflect.Type genericSuperclass = cls.getGenericSuperclass();
    //
    //            if (genericSuperclass != null) {
    //                typeArgs = ((ParameterizedType) genericSuperclass).getActualTypeArguments();
    //            }
    //        }
    //
    //        if (N.notNullOrEmpty(typeArgs)) {
    //            Class<?>[] clses = new Class[typeArgs.length];
    //
    //            for (int i = 0; i < typeArgs.length; i++) {
    //                clses[i] = (Class<?>) typeArgs[i];
    //            }
    //
    //            return clses;
    //        } else {
    //            return null;
    //        }
    //    }
    //

    /**
     * Constructor for N.
     */
    private N() {
        // no instance();
    }

    public static Class<?> forClass(String clsName) {
        return forClass(clsName, true);
    }

    /**
     * support primary type: boolean, char, byte, short, int, long, float,
     * double. and array type with format {@code java.lang.String[]}
     *
     * @param clsName
     * @return
     *
     */
    public static Class<?> forClass(String clsName, boolean cacheResult) {
        Class<?> cls = clsNamePool.get(clsName);

        if (cls == null) {
            cls = BUILT_IN_TYPE.get(clsName);

            if (cls == null) {
                try {
                    cls = Class.forName(clsName);
                } catch (ClassNotFoundException e) {
                    String newClassName = clsName;

                    if (newClassName.indexOf('.') < 0) {
                        int index = newClassName.indexOf("[]");

                        if (((index < 0) && !SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.containsKey(newClassName))
                                || ((index > 0)
                                && !SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.containsKey(newClassName.substring(0, index)))) {
                            newClassName = "java.lang." + newClassName;

                            try {
                                cls = Class.forName(newClassName);

                                BUILT_IN_TYPE.put(clsName, cls);
                            } catch (ClassNotFoundException e1) {
                                // ignore.
                            }
                        }
                    }

                    if (cls == null) {
                        int index = newClassName.indexOf("[]");

                        if (index > 0) {
                            String componentTypeName = newClassName.substring(0, index);
                            String temp = newClassName.replaceAll("\\[\\]", "");

                            if (componentTypeName.equals(temp)) {
                                String symbolOfPrimaryArraryClassName = SYMBOL_OF_PRIMARY_ARRAY_CLASS_NAME.get(componentTypeName);

                                int dimensions = (newClassName.length() - temp.length()) / 2;
                                String prefixOfArray = "";

                                while (dimensions-- > 0) {
                                    prefixOfArray += "[";
                                }

                                if (symbolOfPrimaryArraryClassName != null) {
                                    try {
                                        cls = Class.forName(prefixOfArray + symbolOfPrimaryArraryClassName);

                                        BUILT_IN_TYPE.put(clsName, cls);
                                    } catch (ClassNotFoundException e2) {
                                        // ignore.
                                    }
                                } else {
                                    try {
                                        cls = Class.forName(prefixOfArray + "L" + componentTypeName + ";");
                                    } catch (ClassNotFoundException e3) {
                                        // ignore.
                                    }
                                }
                            }
                        }
                    }
                }
            }

            if (cls == null) {
                throw new RuntimeException("No class found by name: " + clsName);
            }

            if (cacheResult) {
                clsNamePool.put(clsName, cls);
            }
        }

        return cls;
    }

    public static String getClassName(Class<?> cls) {
        String clsName = nameClassPool.get(cls);

        if (clsName == null) {
            clsName = cls.getName();
            nameClassPool.put(cls, clsName);
        }

        return clsName;
    }

    public static String getClassSimpleName(Class<?> cls) {
        String clsName = simpleNameClassPool.get(cls);

        if (clsName == null) {
            clsName = cls.getSimpleName();
            simpleNameClassPool.put(cls, clsName);
        }

        return clsName;
    }

    public static String getClassCanonicalName(Class<?> cls) {
        String clsName = canonicalNameClassPool.get(cls);

        if (clsName == null) {
            clsName = cls.getCanonicalName();
            canonicalNameClassPool.put(cls, clsName);
        }

        return clsName;
    }

    public static Class<?> getEnclosingClass(Class<?> cls) {
        Class<?> enclosingClass = enclosingClassPool.get(cls);

        if (enclosingClass == null) {
            enclosingClass = cls.getEnclosingClass();

            if (enclosingClass == null) {
                enclosingClass = DUMMY_CLASS;
            }

            enclosingClassPool.put(cls, enclosingClass);
        }

        return (enclosingClass == DUMMY_CLASS) ? null : enclosingClass;
    }

    /**
     * Returns the constructor declared in the specified {@code cls} according
     * to the specified {@code parameterType}. {@code null} is returned if no
     * constructor is found.
     *
     * @param cls
     * @param parameterType
     * @return null if constructor not found
     * @throws AbacusException
     */
    public static <T> Constructor<T> getDeclaredConstructor(Class<T> cls, Class<?>... parameterType) {
        try {
            Constructor<T> constructor = cls.getDeclaredConstructor(parameterType);
            constructor.setAccessible(true);

            return constructor;
        } catch (NoSuchMethodException e) {
            return null;
        }
    }

    /**
     * Returns the method declared in the specified {@code cls} according to the
     * specified {@code parameterType}. {@code null} is returned if method not
     * found.
     *
     * @param cls
     *            Class
     * @param methodName
     *            is case insensitive
     * @param parameterTypes
     *            Class<?>[]
     * @return {@code null} if method not found
     * @throws AbacusException
     */
    public static Method getDeclaredMethod(Class<?> cls, String methodName, Class<?>... parameterTypes) {
        Method method = null;

        try {
            method = cls.getDeclaredMethod(methodName, parameterTypes);
        } catch (NoSuchMethodException e) {
            // ignore.
        }

        if (method == null) {
            Method[] methods = cls.getDeclaredMethods();

            for (Method m : methods) {
                if (m.getName().equalsIgnoreCase(methodName) && N.equals(parameterTypes, m.getParameterTypes())) {
                    method = m;

                    break;
                }
            }
        }

        if (method != null) {
            method.setAccessible(true);
        }

        return method;
    }

    public static Method findMethodByName(Class<?> cls, String methodName) {
        Method method = null;

        Method[] methods = cls.getDeclaredMethods();

        for (Method m : methods) {
            if (m.getName().equalsIgnoreCase(methodName)) {
                if ((method == null) || Modifier.isPublic(m.getModifiers())
                        || (Modifier.isProtected(m.getModifiers()) && Modifier.isPrivate(method.getModifiers()))) {
                    method = m;
                }

                if (Modifier.isPublic(method.getModifiers())) {
                    break;
                }
            }
        }

        if (method != null) {
            method.setAccessible(true);
        }

        return method;
    }

    public static <T> T invokeConstructor(Constructor<T> constructor, Object... args) {
        try {
            return (T) constructor.newInstance(args);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        } catch (IllegalArgumentException e) {
            throw new RuntimeException(e);
        } catch (InstantiationException e) {
            throw new RuntimeException(e);
        } catch (InvocationTargetException e) {
            throw new RuntimeException(e);
        }
    }

    public static <T> T invokeMethod(Object instance, Method method, Object... args) {
        try {
            return (T) method.invoke(instance, args);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        } catch (InvocationTargetException e) {
            throw new RuntimeException(e);
        }
    }

    public static String getPropNameByMethod(Method getSetMethod) {
        String propName = methodPropNamePool.get(getSetMethod);

        if (propName == null) {
            String methodName = getSetMethod.getName();
            propName = formalizePropName(methodName.startsWith(N.IS) ? methodName.substring(2) : methodName.substring(3));
            methodPropNamePool.put(getSetMethod, propName);
        }

        return propName;
    }

    public static String formalizePropName(String propName) {
        String newPropName = formalizedPropNamePool.get(propName);

        if (newPropName == null) {
            newPropName = propName;

            for (int i = 0; i < newPropName.length(); i++) {
                if (Character.isLowerCase(newPropName.charAt(i))) {
                    if (i == 1) {
                        newPropName = N.lowCaseFirstChar(newPropName);
                    } else if (i > 1) {
                        newPropName = newPropName.substring(0, i - 1).toLowerCase() + newPropName.substring(i - 1);
                    }

                    break;
                } else if ((i + 1) == newPropName.length()) {
                    newPropName = newPropName.toLowerCase();
                }
            }

            for (String keyWord : keyWords.keySet()) {
                if (keyWord.equalsIgnoreCase(newPropName)) {
                    newPropName = keyWords.get(keyWord);

                    break;
                }
            }

            formalizedPropNamePool.put(propName, newPropName);
        }

        return newPropName;
    }

    public static void registerXMLBindingClassForPropGetSetMethod(Class<?> cls) {
        if (registeredXMLBindingClassList.containsKey(cls)) {
            return;
        }

        synchronized (entityDeclaredPropGetMethodList) {
            registeredXMLBindingClassList.put(cls, false);

            if (entityDeclaredPropGetMethodList.containsKey(cls)) {
                entityDeclaredPropGetMethodList.remove(cls);
                entityDeclaredPropSetMethodList.remove(cls);

                loadPropGetSetMethodList(cls);
            }
        }
    }

    public static void registerPropGetSetMethod(String propName, Method method) {
        synchronized (entityDeclaredPropGetMethodList) {
            Class<?> cls = method.getDeclaringClass();

            if (N.isGetMethod(cls, method)) {
                Map<String, Method> propMethodMap = entityPropGetMethodPool.get(cls);

                if (propMethodMap == null) {
                    loadPropGetSetMethodList(cls);
                    propMethodMap = entityPropGetMethodPool.get(cls);
                }

                if (propMethodMap.containsKey(propName)) {
                    return;
                }

                propMethodMap.put(propName, method);
            } else if (method.getName().startsWith(SET)) {
                Map<String, Method> propMethodMap = entityPropSetMethodPool.get(cls);

                if (propMethodMap == null) {
                    loadPropGetSetMethodList(cls);
                    propMethodMap = entityPropSetMethodPool.get(cls);
                }

                if (propMethodMap.containsKey(propName)) {
                    return;
                }

                if ((method.getParameterTypes().length != 1)) {
                    throw new RuntimeException("Invalid set method: " + method.getName());
                }

                propMethodMap.put(propName, method);
            } else {
                throw new RuntimeException("The name of property getter/setter method must start with 'get' or 'set': "
                    + method.getName());
            }
        }
    }

    /**
     * Call registerClassForJAXBPropGetSetMethod first to retrieve the property getter/setter method for the class/bean generated/wrote by JAXB specification.
     *
     * @param cls
     * @return
     */
    public static Map<String, Method> getPropGetMethodList(Class<?> cls) {
        //        Boolean b = jaxbClassPool.get(cls);
        //
        //        if ((b != null) && b.equals(Boolean.FALSE)) {
        //            loadPropGetSetMethodList(cls);
        //        }
        //
        Map<String, Method> getterMethodList = entityDeclaredPropGetMethodList.get(cls);

        if (getterMethodList == null) {
            loadPropGetSetMethodList(cls);
            getterMethodList = entityDeclaredPropGetMethodList.get(cls);
        }

        return getterMethodList;
    }

    public static Map<String, Method> getPropSetMethodList(Class<?> cls) {
        Map<String, Method> setterMethodList = entityDeclaredPropSetMethodList.get(cls);

        if (setterMethodList == null) {
            loadPropGetSetMethodList(cls);
            setterMethodList = entityDeclaredPropSetMethodList.get(cls);
        }

        return setterMethodList;
    }

    /**
    * Returns the property get method declared in the specified {@code cls}
    * with the specified property name {@code propName} {@code null} is
    * returned if method not found.
    *
    * Call registerClassForJAXBPropGetSetMethod first to retrieve the property getter/setter method for the class/bean generated/wrote by JAXB specification.
    *
    * @param cls
    * @param propName
    * @param isSetByGet
    * @return
    * @throws AbacusException
    */
    public static Method getPropGetMethod(Class<?> cls, String propName) {
        Map<String, Method> propGetMethodMap = entityPropGetMethodPool.get(cls);

        if (propGetMethodMap == null) {
            loadPropGetSetMethodList(cls);
            propGetMethodMap = entityPropGetMethodPool.get(cls);
        }

        Method method = propGetMethodMap.get(propName);

        if (method == null) {
            synchronized (entityDeclaredPropGetMethodList) {
                Map<String, Method> declaredpropGetMethodMap = getPropGetMethodList(cls);

                for (String key : declaredpropGetMethodMap.keySet()) {
                    if (isPropName(cls, propName, key)) {
                        method = declaredpropGetMethodMap.get(key);

                        break;
                    }
                }

                if ((method == null) && !propName.equalsIgnoreCase(N.formalizePropName(propName))) {
                    method = getPropGetMethod(cls, N.formalizePropName(propName));
                }

                // set dummy method t avoid query next time.
                if (method == null) {
                    method = DUMMY_METHOD;
                }

                propGetMethodMap.put(propName, method);
            }
        }

        return (method == DUMMY_METHOD) ? null : method;
    }

    private static boolean isPropName(Class<?> cls, String inputPropName, String propName) {
        return inputPropName.equalsIgnoreCase(propName)
        || inputPropName.replace("_", N.EMPTY_STRING).equalsIgnoreCase(propName)
        || inputPropName.equalsIgnoreCase(N.getClassSimpleName(cls) + '.' + propName.toLowerCase())
        || inputPropName.equalsIgnoreCase(N.getClassSimpleName(cls) + ':' + propName.toLowerCase())
        || (inputPropName.startsWith(N.GET) && inputPropName.substring(3).equalsIgnoreCase(propName))
        || (inputPropName.startsWith(N.SET) && inputPropName.substring(3).equalsIgnoreCase(propName))
        || (inputPropName.startsWith(N.IS) && inputPropName.substring(2).equalsIgnoreCase(propName))
        || (inputPropName.startsWith(N.HAS) && inputPropName.substring(2).equalsIgnoreCase(propName));
    }

    /**
     * Returns the property set method declared in the specified {@code cls}
     * with the specified property name {@code propName} {@code null} is
     * returned if method not found.
     *
     * @param cls
     * @param propName
     * @param isSetByGet
     * @return
     * @throws AbacusException
     */
    public static Method getPropSetMethod(Class<?> cls, String propName) {
        Map<String, Method> propSetMethodMap = entityPropSetMethodPool.get(cls);

        if (propSetMethodMap == null) {
            loadPropGetSetMethodList(cls);
            propSetMethodMap = entityPropSetMethodPool.get(cls);
        }

        Method method = propSetMethodMap.get(propName);

        if (method == null) {
            synchronized (entityDeclaredPropGetMethodList) {
                Map<String, Method> declaredpropSetMethodMap = N.getPropSetMethodList(cls);

                for (String key : declaredpropSetMethodMap.keySet()) {
                    if (isPropName(cls, propName, key)) {
                        method = propSetMethodMap.get(key);

                        break;
                    }
                }

                if ((method == null) && !propName.equalsIgnoreCase(N.formalizePropName(propName))) {
                    method = getPropSetMethod(cls, N.formalizePropName(propName));
                }

                // set dummy method t avoid query next time.
                if (method == null) {
                    method = DUMMY_METHOD;
                }

                propSetMethodMap.put(propName, method);
            }
        }

        return (method == DUMMY_METHOD) ? null : method;
    }

    /**
     * Set the specified {@code propValue} to {@code entity} by the specified
     * method {@code propSetMethod}. This method will try to convert
     * {@code propValue} to appropriate type and set again if fails to set in
     * the first time. The final value which is set to the property will be
     * returned if property is set successfully finally. it could be the input
     * {@code propValue} or converted property value, otherwise, exception will
     * be threw if the property value is set unsuccessfully.
     * @param entity
     *            MapEntity is not supported
     * @param propSetMethod
     * @param propValue
     */
    public static Object setPropValue(Object entity, Method propSetMethod, Object propValue) {
        try {
            propSetMethod.invoke(entity, propValue);
        } catch (Exception e) {
            Class<?> propClass = propSetMethod.getParameterTypes()[0];

            if (propValue == null) {
                propValue = getDefaultValue(propClass);
            }

            try {
                propSetMethod.invoke(entity, propValue);
            } catch (Exception e1) {
                throw new RuntimeException("Failed to set property value by method '" + propSetMethod.getName() + "'", e);
            }
        }

        return propValue;
    }

    public static Object setPropValueByGet(Object entity, Method propGetMethod, Object propValue) {
        Object rt = N.invokeMethod(entity, propGetMethod);

        if (propValue == null) {
            return rt;
        }

        if (rt instanceof Collection) {
            Collection c = (Collection) rt;
            c.addAll((Collection) propValue);
        } else if (rt instanceof Map) {
            Map m = (Map) rt;
            m.putAll((Map) propValue);
        } else {
            throw new RuntimeException("Failed to set property value by method '" + propGetMethod.getName() + "'");
        }

        return rt;
    }

    /**
     * Return the specified {@code propValue} got by the specified method
     * {@code propSetMethod} in the specified {@code entity}.
     * @param entity
     *            MapEntity is not supported
     * @param propGetMethod
     *
     * @return
     */
    public static <T> T getPropValue(Object entity, Method propGetMethod) {
        try {
            return (T) propGetMethod.invoke(entity);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        } catch (InvocationTargetException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Refer to setPropValue(Method, Object, Object).
     *
     * @param entity
     * @param propName
     *            is case insensitive
     * @param propValue
     * @return {@link #setPropValue(Object, Method, Object)}
     */
    public static Object setPropValue(Object entity, String propName, Object propValue) {
        Method propSetMethod = getPropSetMethod(entity.getClass(), propName);

        if (propSetMethod == null) {
            Method propGetMethod = getPropGetMethod(entity.getClass(), propName);

            if (propGetMethod == null) {
                throw new IllegalArgumentException("no setter method found in target class("
                    + N.getClassCanonicalName(entity.getClass()) + ") for property: " + propName);
            } else {
                return setPropValueByGet(entity, propGetMethod, propValue);
            }
        } else {
            return setPropValue(entity, propSetMethod, propValue);
        }
    }

    /**
     * Refer to getPropValue(Method, Object)
     *
     * @param entity
     * @param propName
     *            is case insensitive
     * @return {@link #getPropValue(Object, Method)}
     */
    public static <T> T getPropValue(Object entity, String propName) {
        return getPropValue(entity, getPropGetMethod(entity.getClass(), propName));
    }

    public static <T> T getDefaultValue(Class<T> cls) {
        return (T) DEFAULT_VALUE_FOR_PRIMARY_TYPE.get(cls);
    }

    /**
     * Judged by {@code Double.valueOf(st)}
     *
     * @param value
     * @return
     */
    public static boolean isDigit(String st) {
        try {
            Double.valueOf(st);

            return true;
        } catch (NumberFormatException e) {
            return false;
        }
    }

    public static boolean isNullOrEmpty(String s) {
        return (s == null) || (s.length() == 0);
    }

    public static boolean isNullOrEmpty(Collection<?> c) {
        return (c == null) || (c.size() == 0);
    }

    public static boolean isNullOrEmpty(Map<?, ?> m) {
        return (m == null) || (m.size() == 0);
    }

    public static boolean isNullOrEmpty(boolean[] a) {
        return (a == null) || (a.length == 0);
    }

    public static boolean isNullOrEmpty(char[] a) {
        return (a == null) || (a.length == 0);
    }

    public static boolean isNullOrEmpty(byte[] a) {
        return (a == null) || (a.length == 0);
    }

    public static boolean isNullOrEmpty(short[] a) {
        return (a == null) || (a.length == 0);
    }

    public static boolean isNullOrEmpty(int[] a) {
        return (a == null) || (a.length == 0);
    }

    public static boolean isNullOrEmpty(long[] a) {
        return (a == null) || (a.length == 0);
    }

    public static boolean isNullOrEmpty(float[] a) {
        return (a == null) || (a.length == 0);
    }

    public static boolean isNullOrEmpty(double[] a) {
        return (a == null) || (a.length == 0);
    }

    public static boolean isNullOrEmpty(Object[] a) {
        return (a == null) || (a.length == 0);
    }

    public static boolean isNullOrEmpty(NodeList nodeList) {
        return (nodeList == null) || (nodeList.getLength() == 0);
    }

    public static boolean notNullOrEmpty(String s) {
        return (s != null) && (s.length() > 0);
    }

    public static boolean notNullOrEmpty(Collection<?> c) {
        return (c != null) && (c.size() > 0);
    }

    public static boolean notNullOrEmpty(Map<?, ?> m) {
        return (m != null) && (m.size() > 0);
    }

    public static boolean notNullOrEmpty(boolean[] a) {
        return (a != null) && (a.length > 0);
    }

    public static boolean notNullOrEmpty(char[] a) {
        return (a != null) && (a.length > 0);
    }

    public static boolean notNullOrEmpty(byte[] a) {
        return (a != null) && (a.length > 0);
    }

    public static boolean notNullOrEmpty(short[] a) {
        return (a != null) && (a.length > 0);
    }

    public static boolean notNullOrEmpty(int[] a) {
        return (a != null) && (a.length > 0);
    }

    public static boolean notNullOrEmpty(long[] a) {
        return (a != null) && (a.length > 0);
    }

    public static boolean notNullOrEmpty(float[] a) {
        return (a != null) && (a.length > 0);
    }

    public static boolean notNullOrEmpty(double[] a) {
        return (a != null) && (a.length > 0);
    }

    public static boolean notNullOrEmpty(Object[] a) {
        return (a != null) && (a.length > 0);
    }

    public static boolean notNullOrEmpty(NodeList nodeList) {
        return (nodeList != null) && (nodeList.getLength() > 0);
    }

    private static void loadPropGetSetMethodList(Class<?> cls) {
        synchronized (entityDeclaredPropGetMethodList) {
            if (entityDeclaredPropGetMethodList.containsKey(cls)) {
                return;
            }

            Object instance = null;

            if (registeredXMLBindingClassList.containsKey(cls)) {
                try {
                    instance = cls.newInstance();
                } catch (Exception e) {
                    logger.warning("Failed to new instance of class: " + cls.getCanonicalName()
                        + " to check setter method by getter method");
                }

                registeredXMLBindingClassList.put(cls, true);
            }

            Map<String, Method> propGetMethodMap = N.asLinkedHashMap();
            Map<String, Method> propSetMethodMap = N.asLinkedHashMap();

            List<Class<?>> allClasses = N.asList();
            allClasses.add(cls);

            while (allClasses.get(allClasses.size() - 1).getSuperclass() != null) {
                allClasses.add(allClasses.get(allClasses.size() - 1).getSuperclass());
            }

            Class<?> clazz = null;
            Method setMethod = null;

            for (int i = allClasses.size() - 1; i >= 0; i--) {
                clazz = allClasses.get(i);

                if (java.util.Date.class.isAssignableFrom(clazz) || Calendar.class.equals(clazz)) {
                    continue;
                }

                Map<String, String> statisFinalFields = getPublicStaticStringFields(clazz);

                String propName = null;

                // sort the methods by the order of declared fields
                String fieldName = null;

                for (Field field : clazz.getDeclaredFields()) {
                    fieldName = field.getName();

                    for (Method method : clazz.getMethods()) {
                        if (isFieldGetMethod(clazz, method, fieldName)) {
                            propName = getPropNameByMethod(method);
                            propName = (statisFinalFields.get(propName) != null) ? statisFinalFields.get(propName)
                                                                                 : propName;

                            if (propGetMethodMap.containsKey(propName)) {
                                break;
                            }

                            setMethod = getSetMethod(clazz, propName, method);

                            if (setMethod != null) {
                                propGetMethodMap.put(propName, method);
                                propSetMethodMap.put(propName, setMethod);

                                break;
                            }

                            if (isJAXBGetMethod(instance, method)) {
                                propGetMethodMap.put(propName, method);

                                break;
                            }
                        }
                    }
                }

                for (Method method : clazz.getMethods()) {
                    if (isGetMethod(clazz, method)) {
                        propName = getPropNameByMethod(method);
                        propName = (statisFinalFields.get(propName) != null) ? statisFinalFields.get(propName) : propName;

                        if (propGetMethodMap.containsKey(propName)) {
                            break;
                        }

                        setMethod = getSetMethod(clazz, propName, method);

                        if (setMethod != null) {
                            propGetMethodMap.put(propName, method);
                            propSetMethodMap.put(propName, setMethod);

                            break;
                        }

                        if (isJAXBGetMethod(instance, method)) {
                            propGetMethodMap.put(propName, method);

                            break;
                        }
                    }
                }
            }

            // for Double-Checked Locking is Broken. initialize it before
            // put it into map.
            Map<String, Method> temp = Collections.unmodifiableMap(propGetMethodMap);
            temp.keySet();
            entityDeclaredPropGetMethodList.put(cls, temp);

            if (entityPropGetMethodPool.get(cls) == null) {
                entityPropGetMethodPool.put(cls, N.asConcurrentHashMap(propGetMethodMap));
            } else {
                entityPropGetMethodPool.get(cls).putAll(propGetMethodMap);
            }

            // for Double-Checked Locking is Broken. initialize it before
            // put it into map.
            temp = Collections.unmodifiableMap(propSetMethodMap);
            temp.keySet();
            entityDeclaredPropSetMethodList.put(cls, temp);

            if (entityPropSetMethodPool.get(cls) == null) {
                entityPropSetMethodPool.put(cls, N.asConcurrentHashMap(propSetMethodMap));
            } else {
                entityPropSetMethodPool.get(cls).putAll(propSetMethodMap);
            }
        }
    }

    private static boolean isGetMethod(Class<?> clazz, Method method) {
        //        if (DirtyMarker.class.isAssignableFrom(clazz)
        //                && (method.getName().equals("getSignedPropNames") || method.getName().equals("getModifiedPropNames"))) {
        //            return false;
        //        }
        //
        return (method.getName().startsWith(GET) || method.getName().startsWith(IS) || method.getName().startsWith(HAS))
        && (method.getParameterTypes().length == 0) && !void.class.equals(method.getReturnType());
    }

    private static boolean isFieldGetMethod(Class<?> clazz, Method method, String fieldName) {
        return isGetMethod(clazz, method)
        && (method.getName().equalsIgnoreCase(GET + fieldName) || method.getName().equalsIgnoreCase(IS + fieldName)
        || method.getName().equalsIgnoreCase(HAS + fieldName));
    }

    private static boolean isJAXBGetMethod(Object instance, Method method) {
        try {
            return (instance != null)
            && (Collection.class.isAssignableFrom(method.getReturnType())
            || Map.class.isAssignableFrom(method.getReturnType())) && (N.invokeMethod(instance, method) != null);
        } catch (Exception e) {
            return false;
        }
    }

    private static Method getSetMethod(Class<?> clazz, String propName, Method getMethod) {
        Method setMethod = N.getDeclaredMethod(clazz, N.SET + propName, getMethod.getReturnType());

        return ((setMethod == null) || !void.class.equals(setMethod.getReturnType())) ? null : setMethod;
    }

    private static <T> Map<String, String> getPublicStaticStringFields(Class<T> cls) {
        Map<String, String> statisFinalFields = N.asMap();

        for (Field field : cls.getFields()) {
            if (Modifier.isPublic(field.getModifiers()) && Modifier.isStatic(field.getModifiers())
                    && Modifier.isFinal(field.getModifiers()) && String.class.equals(field.getType())) {
                String value;

                try {
                    value = (String) field.get(null);
                    statisFinalFields.put(value, value);
                } catch (IllegalAccessException e) {
                    // ignore. should not happen.
                }
            }
        }

        return statisFinalFields;
    }

    public static void erase(Object entity) {
        if (entity == null) {
            return;
        }

        Class<?> cls = entity.getClass();
        Map<String, Method> setterMethodList = N.getPropSetMethodList(cls);

        if (setterMethodList.size() == 0) {
            throw new IllegalArgumentException("no property getXXX/setXXX method found in the specified entity: "
                + N.getClassCanonicalName(cls));
        }

        for (Method propSetMethod : setterMethodList.values()) {
            N.setPropValue(entity, propSetMethod, null);
        }
    }

    /**
     * Set the non-null properties(including all primary type properties) in the
     * specified {@code sourceEntity} to the specified {@code targetEntity}.
     *
     * @param sourceEntity
     *            a Java Object what allows access to properties using getter
     *            and setter methods.
     * @param targetEntity
     *            a Java Object what allows access to properties using getter
     *            and setter methods.
     */
    public static void merge(Object sourceEntity, Object targetEntity) {
        Class<?> sourceEntityClass = sourceEntity.getClass();
        Class<?> targetEntityClass = targetEntity.getClass();

        Map<String, Method> getterMethodList = N.getPropGetMethodList(sourceEntityClass);

        if (getterMethodList.size() == 0) {
            throw new IllegalArgumentException("no property getXXX/setXXX method found in the specified entity: "
                + sourceEntityClass.getCanonicalName());
        }

        Method propSetMethod = null;
        Method propGetMethod = null;
        String propName = null;
        Object propValue = null;

        try {
            for (Method srcPropGetMethod : getterMethodList.values()) {
                propName = getPropNameByMethod(srcPropGetMethod);
                propValue = srcPropGetMethod.invoke(sourceEntity);
                propSetMethod = getPropSetMethod(targetEntityClass, propName);

                if (propSetMethod == null) {
                    propGetMethod = getPropGetMethod(targetEntityClass, propName);

                    if (propGetMethod == null) {
                        throw new IllegalArgumentException("no setter method found in target class("
                            + N.getClassCanonicalName(targetEntityClass) + ") for property: " + propName);
                    } else {
                        if (propValue != null) {
                            setPropValueByGet(targetEntity, propGetMethod, propValue);
                        }
                    }
                } else {
                    if (propValue != null) {
                        setPropValue(targetEntity, propSetMethod, propValue);
                    }
                }
            }
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        } catch (InvocationTargetException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Returns a new created instance of the same class and set with same
     * properties retrieved by 'getXXX' method in the specified {@code entity}.
     *
     * @param entity
     *            a Java Object what allows access to properties using getter
     *            and setter methods.
     * @return
     */
    public static <T> T copy(T entity) {
        return (T) copy(entity.getClass(), entity);
    }

    /**
     * Returns a new created instance of the specified {@code cls} and set with
     * same properties retrieved by 'getXXX' method in the specified
     * {@code entity}.
     *
     * @param cls
     *            a Java Object what allows access to properties using getter
     *            and setter methods.
     * @param entity
     *            a Java Object what allows access to properties using getter
     *            and setter methods.
     * @return
     */
    public static <T> T copy(Class<T> cls, Object entity) {
        Class<?> srcCls = entity.getClass();
        Object copy = newInstance(cls);

        Map<String, Method> srcGetterMethodList = N.getPropGetMethodList(srcCls);

        if (srcGetterMethodList.size() == 0) {
            throw new IllegalArgumentException("no property getXXX/setXXX method found in the specified entity: "
                + entity.getClass().getCanonicalName());
        }

        try {
            Method propGetMethod = null;
            Method propSetMethod = null;
            String propName = null;

            for (Method srcPropGetMethod : srcGetterMethodList.values()) {
                propName = getPropNameByMethod(srcPropGetMethod);
                propSetMethod = N.getPropSetMethod(cls, propName);

                if (propSetMethod == null) {
                    propGetMethod = N.getPropGetMethod(cls, propName);

                    if (propGetMethod == null) {
                        throw new IllegalArgumentException("no setter method found in target class("
                            + N.getClassCanonicalName(cls) + ") for property: " + propName);
                    } else {
                        setPropValueByGet(copy, propGetMethod, srcPropGetMethod.invoke(entity));
                    }
                } else {
                    setPropValue(copy, propSetMethod, srcPropGetMethod.invoke(entity));
                }
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

        return (T) copy;
    }

    public static <T> T map2Entity(Class<T> cls, Map<String, Object> m) {
        return map2Entity(cls, m, true, true);
    }

    public static <T> T map2Entity(Class<T> cls, Map<String, Object> m, boolean ignoreUnknownProperty,
        boolean ignoreNullProperty) {
        checkEntityClass(cls);

        T entity = newInstance(cls);
        Method propGetMethod = null;
        Method propSetMethod = null;

        Object propValue = null;

        for (String propName : m.keySet()) {
            propValue = m.get(propName);
            propSetMethod = getPropSetMethod(cls, propName);

            if (propSetMethod == null) {
                propGetMethod = getPropGetMethod(cls, propName);
            }

            if ((propSetMethod == null) && (propGetMethod == null)) {
                if (ignoreUnknownProperty) {
                    continue;
                } else {
                    throw new RuntimeException("unknown proprety: " + propName);
                }
            }

            if ((propValue == null) && ignoreNullProperty) {
                continue;
            }

            if (propSetMethod == null) {
                propGetMethod = getPropGetMethod(cls, propName);

                if (propGetMethod != null) {
                    setPropValueByGet(entity, propGetMethod, propValue);
                }
            } else {
                if ((propValue != null) && (N.getClazzParserType(propValue.getClass()) == ClazzParserType.MAP)
                        && (N.getClazzParserType(propSetMethod.getParameterTypes()[0]) != ClazzParserType.MAP)) {
                    setPropValue(entity, propSetMethod,
                        map2Entity(propSetMethod.getParameterTypes()[0], (Map<String, Object>) propValue,
                            ignoreUnknownProperty, ignoreNullProperty));
                } else {
                    setPropValue(entity, propSetMethod, propValue);
                }
            }
        }

        return entity;
    }

    public static <T> List<T> map2Entity(Class<T> cls, List<Map<String, Object>> mList) {
        return map2Entity(cls, mList, true, true);
    }

    public static <T> List<T> map2Entity(Class<T> cls, List<Map<String, Object>> mList, boolean ignoreUnknownProperty,
        boolean igoreNullProperty) {
        checkEntityClass(cls);

        List<T> entityList = N.newArrayList(mList.size());

        for (Map<String, Object> m : mList) {
            entityList.add(map2Entity(cls, m, ignoreUnknownProperty, igoreNullProperty));
        }

        return entityList;
    }

    public static Map<String, Object> entity2Map(Object entity) {
        return entity2Map(entity, true);
    }

    public static Map<String, Object> entity2Map(Object entity, boolean ignoreNullProperty) {
        Map<String, Object> resultMap = N.asMap();
        entity2Map(entity, resultMap, ignoreNullProperty);

        return resultMap;
    }

    public static void entity2Map(Object entity, Map<String, Object> resultMap) {
        entity2Map(entity, resultMap, true);
    }

    public static void entity2Map(Object entity, Map<String, Object> resultMap, boolean ignoreNullProperty) {
        Map<String, Method> getterMethodList = N.getPropGetMethodList(entity.getClass());

        if (getterMethodList.size() == 0) {
            throw new IllegalArgumentException("no property getXXX/setXXX method found in the specified entity: "
                + entity.getClass().getCanonicalName());
        }

        try {
            Method propGetMethod = null;
            Object propValue = null;

            for (String propName : getterMethodList.keySet()) {
                propGetMethod = getterMethodList.get(propName);
                propValue = propGetMethod.invoke(entity);

                if ((propValue == null) && ignoreNullProperty) {
                    continue;
                }

                resultMap.put(getPropNameByMethod(propGetMethod), propValue);
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static List<Map<String, Object>> entity2Map(List<?> entityList) {
        return entity2Map(entityList, true);
    }

    public static List<Map<String, Object>> entity2Map(List<?> entityList, boolean ignoreNullProperty) {
        List<Map<String, Object>> mList = N.newArrayList(entityList.size());

        for (Object entity : entityList) {
            mList.add(entity2Map(entity, ignoreNullProperty));
        }

        return mList;
    }

    public static Map<String, Object> deepEntity2Map(Object entity) {
        return deepEntity2Map(entity, true);
    }

    public static Map<String, Object> deepEntity2Map(Object entity, boolean ignoreNullProperty) {
        Map<String, Object> resultMap = N.asMap();
        deepEntity2Map(entity, resultMap, ignoreNullProperty);

        return resultMap;
    }

    public static void deepEntity2Map(Object entity, Map<String, Object> resultMap) {
        deepEntity2Map(entity, resultMap, true);
    }

    public static void deepEntity2Map(Object entity, Map<String, Object> resultMap, boolean ignoreNullProperty) {
        Map<String, Method> getterMethodList = N.getPropGetMethodList(entity.getClass());

        if (getterMethodList.size() == 0) {
            throw new IllegalArgumentException("no property getXXX/setXXX method found in the specified entity: "
                + entity.getClass().getCanonicalName());
        }

        try {
            Method propGetMethod = null;
            Object propValue = null;

            for (String propName : getterMethodList.keySet()) {
                propGetMethod = getterMethodList.get(propName);
                propValue = propGetMethod.invoke(entity);

                if ((propValue == null) && ignoreNullProperty) {
                    continue;
                }

                if ((propValue == null) || (N.getPropGetMethodList(propValue.getClass()).size() == 0)) {
                    resultMap.put(propName, propValue);
                } else {
                    resultMap.put(propName, deepEntity2Map(propValue, ignoreNullProperty));
                }
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static List<Map<String, Object>> deepEntity2Map(List<?> entityList) {
        return deepEntity2Map(entityList, true);
    }

    public static List<Map<String, Object>> deepEntity2Map(List<?> entityList, boolean ignoreNullProperty) {
        List<Map<String, Object>> mList = N.newArrayList(entityList.size());

        for (Object entity : entityList) {
            mList.add(deepEntity2Map(entity, ignoreNullProperty));
        }

        return mList;
    }

    public static Map<String, Object> entity2FlatMap(Object entity) {
        return entity2FlatMap(entity, true);
    }

    public static Map<String, Object> entity2FlatMap(Object entity, boolean ignoreNullProperty) {
        Map<String, Object> resultMap = N.asMap();
        entity2FlatMap(entity, resultMap, ignoreNullProperty);

        return resultMap;
    }

    public static void entity2FlatMap(Object entity, Map<String, Object> resultMap) {
        entity2Map(entity, resultMap, true);
    }

    public static void entity2FlatMap(Object entity, Map<String, Object> resultMap, boolean ignoreNullProperty) {
        Map<String, Method> getterMethodList = N.getPropGetMethodList(entity.getClass());

        if (getterMethodList.size() == 0) {
            throw new IllegalArgumentException("no property getXXX/setXXX method found in the specified entity: "
                + entity.getClass().getCanonicalName());
        }

        try {
            Method propGetMethod = null;
            Object propValue = null;

            for (String propName : getterMethodList.keySet()) {
                propGetMethod = getterMethodList.get(propName);
                propValue = propGetMethod.invoke(entity);

                prop2FlatMap(propName, propValue, resultMap, ignoreNullProperty);
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static List<Map<String, Object>> entity2FlatMap(List<?> entityList) {
        return entity2FlatMap(entityList, true);
    }

    public static List<Map<String, Object>> entity2FlatMap(List<?> entityList, boolean ignoreNullProperty) {
        List<Map<String, Object>> mList = N.newArrayList(entityList.size());

        for (Object entity : entityList) {
            mList.add(entity2FlatMap(entity, ignoreNullProperty));
        }

        return mList;
    }

    private static void prop2FlatMap(String propName, Object propValue, Map<String, Object> resultMap,
        boolean ignoreNullProperty) {
        if (propValue == null) {
            if (!ignoreNullProperty) {
                resultMap.put(propName, propValue);
            }

            return;
        }

        Class<?> cls = propValue.getClass();

        ClazzParserType clsParserType = N.getClazzParserType(cls);

        switch (clsParserType) {
        case SINGLE_VALUE:
            resultMap.put(propName, propValue);

            break;

        case ARRAY:

            Object[] a = (Object[]) propValue;

            for (int i = 0; i < a.length; i++) {
                prop2FlatMap(propName + "[" + i + "]", a[i], resultMap, ignoreNullProperty);
            }

            break;

        case COLLECTION:

            Collection<Object> c = (Collection<Object>) propValue;
            int i = 0;

            for (Object e : c) {
                prop2FlatMap(propName + "[" + i++ + "]", e, resultMap, ignoreNullProperty);
            }

            break;

        case MAP:

            Map<Object, Object> map = (Map) propValue;

            for (Object k : map.keySet()) {
                prop2FlatMap(propName + "." + toString(k), map.get(k), resultMap, ignoreNullProperty);
            }

            break;

        case ENTITY:

            Map<String, Method> getterMethodList = N.getPropGetMethodList(cls);

            try {
                Method propGetMethod = null;
                Object subPropValue = null;

                for (String subPropName : getterMethodList.keySet()) {
                    propGetMethod = getterMethodList.get(subPropName);
                    subPropValue = propGetMethod.invoke(propValue);

                    prop2FlatMap(propName + "." + subPropName, subPropValue, resultMap, ignoreNullProperty);
                }
            } catch (Exception e) {
                throw new RuntimeException(e);
            }

            break;

        default:
            resultMap.put(propName, propValue);
        }
    }

    /**
     * @see Arrays#copyOfRange(boolean[], int, int)
     *
     * @param a
     * @param from
     * @param to
     * @return
     */
    public static boolean[] copy(boolean[] a, int from, int to) {
        return Arrays.copyOfRange(a, from, to);
    }

    /**
     * @see Arrays#copyOfRange(char[], int, int)
     *
     * @param a
     * @param from
     * @param to
     * @return
     */
    public static char[] copy(char[] a, int from, int to) {
        return Arrays.copyOfRange(a, from, to);
    }

    /**
     * @see Arrays#copyOfRange(byte[], int, int)
     *
     * @param a
     * @param from
     * @param to
     * @return
     */
    public static byte[] copy(byte[] a, int from, int to) {
        return Arrays.copyOfRange(a, from, to);
    }

    /**
     * @see Arrays#copyOfRange(short[], int, int)
     *
     * @param a
     * @param from
     * @param to
     * @return
     */
    public static short[] copy(short[] a, int from, int to) {
        return Arrays.copyOfRange(a, from, to);
    }

    /**
     * @see Arrays#copyOfRange(int[], int, int)
     *
     * @param a
     * @param from
     * @param to
     * @return
     */
    public static int[] copy(int[] a, int from, int to) {
        return Arrays.copyOfRange(a, from, to);
    }

    /**
     * @see Arrays#copyOfRange(long[], int, int)
     *
     * @param a
     * @param from
     * @param to
     * @return
     */
    public static long[] copy(long[] a, int from, int to) {
        return Arrays.copyOfRange(a, from, to);
    }

    /**
     * @see Arrays#copyOfRange(float[], int, int)
     *
     * @param a
     * @param from
     * @param to
     * @return
     */
    public static float[] copy(float[] a, int from, int to) {
        return Arrays.copyOfRange(a, from, to);
    }

    /**
     * @see Arrays#copyOfRange(double[], int, int)
     *
     * @param a
     * @param from
     * @param to
     * @return
     */
    public static double[] copy(double[] a, int from, int to) {
        return Arrays.copyOfRange(a, from, to);
    }

    /**
     * @see Arrays#copyOfRange(T[], int, int)
     * @param a
     * @param from
     * @param to
     * @return
     */
    public static <T> T[] copy(T[] a, int from, int to) {
        return Arrays.copyOfRange(a, from, to);
    }

    /**
     * @see System#arraycopy(Object, int, Object, int, int) is called
     *
     * @param src
     * @param srcPos
     * @param dest
     * @param destPos
     * @param length
     * @return the destination array is returned.
     */
    public static <T> T[] copy(T[] src, int srcPos, T[] dest, int destPos, int length) {
        System.arraycopy(src, srcPos, dest, destPos, length);

        return dest;
    }

    /**
     * Returns a new array combined by the two specified array {@code a} and
     * {@code b}.
     *
     * @param a
     * @param b
     * @return
     */
    public static boolean[] combine(boolean[] a, boolean[] b) {
        boolean[] c = new boolean[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);

        return c;
    }

    /**
     * Returns a new array combined by the two specified array {@code a} and
     * {@code b}.
     *
     * @param a
     * @param b
     * @return
     */
    public static char[] combine(char[] a, char[] b) {
        char[] c = new char[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);

        return c;
    }

    /**
     * Returns a new array combined by the two specified array {@code a} and
     * {@code b}.
     *
     * @param a
     * @param b
     * @return
     */
    public static byte[] combine(byte[] a, byte[] b) {
        byte[] c = new byte[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);

        return c;
    }

    /**
     * Returns a new array combined by the two specified array {@code a} and
     * {@code b}.
     *
     * @param a
     * @param b
     * @return
     */
    public static short[] combine(short[] a, short[] b) {
        short[] c = new short[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);

        return c;
    }

    /**
     * Returns a new array combined by the two specified array {@code a} and
     * {@code b}.
     *
     * @param a
     * @param b
     * @return
     */
    public static int[] combine(int[] a, int[] b) {
        int[] c = new int[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);

        return c;
    }

    /**
     * Returns a new array combined by the two specified array {@code a} and
     * {@code b}.
     *
     * @param a
     * @param b
     * @return
     */
    public static long[] combine(long[] a, long[] b) {
        long[] c = new long[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);

        return c;
    }

    /**
     * Returns a new array combined by the two specified array {@code a} and
     * {@code b}.
     *
     * @param a
     * @param b
     * @return
     */
    public static float[] combine(float[] a, float[] b) {
        float[] c = new float[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);

        return c;
    }

    /**
     * Returns a new array combined by the two specified array {@code a} and
     * {@code b}.
     *
     * @param a
     * @param b
     * @return
     */
    public static double[] combine(double[] a, double[] b) {
        double[] c = new double[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);

        return c;
    }

    /**
     * Returns a new array combined by the two specified array {@code a} and
     * {@code b}.
     *
     * @param a
     * @param b
     * @return
     */
    public static <T> T[] combine(T[] a, T[] b) {
        T[] c = (T[]) N.newArray(a.getClass().getComponentType(), a.length + b.length);
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);

        return c;
    }

    public static <T extends Comparable<T>> void sort(T[] a) {
        Arrays.sort(a);
    }

    public static <T> void sort(T[] a, Comparator<?super T> c) {
        Arrays.sort(a, c);
    }

    public static <T extends Comparable<T>> void sort(List<T> list) {
        Collections.sort(list);
    }

    public static <T> void sort(List<T> list, Comparator<?super T> c) {
        Collections.sort(list, c);
    }

    public static int indexOf(boolean e, boolean[] a) {
        for (int i = 0; i < a.length; i++) {
            if (a[i] == e) {
                return i;
            }
        }

        return -1;
    }

    public static int indexOf(char e, char[] a) {
        for (int i = 0; i < a.length; i++) {
            if (a[i] == e) {
                return i;
            }
        }

        return -1;
    }

    public static int indexOf(byte e, byte[] a) {
        for (int i = 0; i < a.length; i++) {
            if (a[i] == e) {
                return i;
            }
        }

        return -1;
    }

    public static int indexOf(short e, short[] a) {
        for (int i = 0; i < a.length; i++) {
            if (a[i] == e) {
                return i;
            }
        }

        return -1;
    }

    public static int indexOf(int e, int[] a) {
        for (int i = 0; i < a.length; i++) {
            if (a[i] == e) {
                return i;
            }
        }

        return -1;
    }

    public static int indexOf(long e, long[] a) {
        for (int i = 0; i < a.length; i++) {
            if (a[i] == e) {
                return i;
            }
        }

        return -1;
    }

    public static int indexOf(float e, float[] a) {
        for (int i = 0; i < a.length; i++) {
            if (Float.compare(a[i], e) == 0) {
                return i;
            }
        }

        return -1;
    }

    public static int indexOf(double e, double[] a) {
        for (int i = 0; i < a.length; i++) {
            if (Double.compare(a[i], e) == 0) {
                return i;
            }
        }

        return -1;
    }

    public static int indexOf(Object e, Object[] a) {
        for (int i = 0; i < a.length; i++) {
            if (equals(a[i], e)) {
                return i;
            }
        }

        return -1;
    }

    /**
     * Method equals.
     *
     * @param a
     *            boolean
     * @param b
     *            boolean
     * @return boolean
     */
    public static boolean equals(boolean a, boolean b) {
        return a == b;
    }

    /**
     * Method equals.
     *
     * @param a
     *            char
     * @param b
     *            char
     * @return boolean
     */
    public static boolean equals(char a, char b) {
        return a == b;
    }

    /**
     * Method equals.
     *
     * @param a
     *            byte
     * @param b
     *            byte
     * @return boolean
     */
    public static boolean equals(byte a, byte b) {
        return a == b;
    }

    /**
     * Method equals.
     *
     * @param a
     *            short
     * @param b
     *            short
     * @return boolean
     */
    public static boolean equals(short a, short b) {
        return a == b;
    }

    /**
     * Method equals.
     *
     * @param a
     *            int
     * @param b
     *            int
     * @return boolean
     */
    public static boolean equals(int a, int b) {
        return a == b;
    }

    /**
     * Method equals.
     *
     * @param a
     *            long
     * @param b
     *            long
     * @return boolean
     */
    public static boolean equals(long a, long b) {
        return a == b;
    }

    /**
     * Method equals.
     *
     * @param a
     *            float
     * @param b
     *            float
     * @return boolean
     */
    public static boolean equals(float a, float b) {
        return Float.compare(a, b) == 0;
    }

    /**
     * Method equals.
     *
     * @param a
     *            double
     * @param b
     *            double
     * @return boolean
     */
    public static boolean equals(double a, double b) {
        return Double.compare(a, b) == 0;
    }

    /**
     *
     * @param a
     * @param b
     * @return
     */
    public static boolean equals(String a, String b) {
        return ((a == null) && (b == null)) || ((a != null) && a.equals(b));
    }

    /**
     * compare {@code a} and {@code b} by
     * {@link Arrays#equals(Object[], Object[])} if both of them are array.
     *
     * @param a
     *            Object
     * @param b
     *            Object
     * @return boolean
     */
    public static boolean equals(Object a, Object b) {
        if ((a != null) && (b != null) && a.getClass().isArray() && a.getClass().equals(b.getClass())) {
            int order = getClazzEnumOrder(a.getClass());

            switch (order) {
            case 1:
                return equals((boolean[]) a, (boolean[]) b);

            case 2:
                return equals((char[]) a, (char[]) b);

            case 3:
                return equals((byte[]) a, (byte[]) b);

            case 4:
                return equals((short[]) a, (short[]) b);

            case 5:
                return equals((int[]) a, (int[]) b);

            case 6:
                return equals((long[]) a, (long[]) b);

            case 7:
                return equals((float[]) a, (float[]) b);

            case 8:
                return equals((double[]) a, (double[]) b);

            default:
                return equals((Object[]) a, (Object[]) b);
            }
        }

        return (a == null) ? (b == null) : a.equals(b);
    }

    /**
     * compare {@code a} and {@code b} by
     * {@link Arrays#equals(Object[], Object[])} if both of them are array.
     *
     * @param a
     *            Object
     * @param b
     *            Object
     * @return boolean
     */
    public static boolean deepEquals(Object a, Object b) {
        if ((a != null) && (b != null) && a.getClass().isArray() && a.getClass().equals(b.getClass())) {
            int order = getClazzEnumOrder(a.getClass());

            switch (order) {
            case 1:
                return equals((boolean[]) a, (boolean[]) b);

            case 2:
                return equals((char[]) a, (char[]) b);

            case 3:
                return equals((byte[]) a, (byte[]) b);

            case 4:
                return equals((short[]) a, (short[]) b);

            case 5:
                return equals((int[]) a, (int[]) b);

            case 6:
                return equals((long[]) a, (long[]) b);

            case 7:
                return equals((float[]) a, (float[]) b);

            case 8:
                return equals((double[]) a, (double[]) b);

            default:
                return deepEquals((Object[]) a, (Object[]) b);
            }
        }

        return (a == null) ? (b == null) : a.equals(b);
    }

    /**
     * @see Arrays#equals(boolean[], boolean[])
     *
     * @param a
     *            boolean[]
     * @param b
     *            boolean[]
     * @return boolean
     */
    public static boolean equals(boolean[] a, boolean[] b) {
        return Arrays.equals(a, b);
    }

    /**
     * @see Arrays#equals(char[], char[])
     *
     * @param a
     *            char[]
     * @param b
     *            char[]
     * @return boolean
     */
    public static boolean equals(char[] a, char[] b) {
        return Arrays.equals(a, b);
    }

    /**
     * @see Arrays#equals(byte[], byte[])
     *
     * @param a
     *            byte[]
     * @param b
     *            byte[]
     * @return boolean
     */
    public static boolean equals(byte[] a, byte[] b) {
        return Arrays.equals(a, b);
    }

    /**
     * @see Arrays#equals(short[], short[])
     * @param a
     * @param b
     * @return
     */
    public static boolean equals(short[] a, short[] b) {
        return Arrays.equals(a, b);
    }

    /**
     * @see Arrays#equals(int[], int[])
     *
     * @param a
     *            int[]
     * @param b
     *            int[]
     * @return boolean
     */
    public static boolean equals(int[] a, int[] b) {
        return Arrays.equals(a, b);
    }

    /**
     * @see Arrays#equals(long[], long[])
     *
     * @param a
     *            long[]
     * @param b
     *            long[]
     * @return boolean
     */
    public static boolean equals(long[] a, long[] b) {
        return Arrays.equals(a, b);
    }

    /**
     * @see Arrays#equals(float[], float[])
     *
     * @param a
     *            float[]
     * @param b
     *            float[]
     * @return boolean
     */
    public static boolean equals(float[] a, float[] b) {
        return Arrays.equals(a, b);
    }

    /**
     * @see Arrays#equals(double[], double[])
     *
     * @param a
     *            double[]
     * @param b
     *            double[]
     * @return boolean
     */
    public static boolean equals(double[] a, double[] b) {
        return Arrays.equals(a, b);
    }

    /**
     * @see Arrays#equals(Object[], Object[])
     * @param a
     *            Object[]
     * @param b
     *            Object[]
     * @return boolean
     */
    public static boolean equals(Object[] a, Object[] b) {
        return Arrays.equals(a, b);
    }

    /**
     * @see Arrays#deepEquals(Object[], Object[])
     * @param a
     * @param b
     * @return
     */
    public static boolean deepEquals(Object[] a, Object[] b) {
        return Arrays.deepEquals(a, b);
    }

    /**
     * Method hashCode.
     *
     * @param value
     *            boolean
     * @return int
     */
    public static int hashCode(boolean value) {
        return value ? 1231 : 1237;
    }

    /**
     * Method hashCode.
     *
     * @param value
     *            char
     * @return int
     */
    public static int hashCode(char value) {
        return (int) value;
    }

    /**
     * Method hashCode.
     *
     * @param value
     *            byte
     * @return int
     */
    public static int hashCode(byte value) {
        return (int) value;
    }

    /**
     * Method hashCode.
     *
     * @param value
     * @return int
     */
    public static int hashCode(short value) {
        return (int) value;
    }

    /**
     * Method hashCode.
     *
     * @param vaue
     *            int
     * @return int
     */
    public static int hashCode(int value) {
        return value;
    }

    /**
     * Method hashCode.
     *
     * @param value
     *            long
     * @return int
     */
    public static int hashCode(long value) {
        return (int) (value ^ (value >>> 32));
    }

    /**
     * Method hashCode.
     *
     * @param value
     *            float
     * @return int
     */
    public static int hashCode(float value) {
        return Float.floatToIntBits(value);
    }

    /**
     * Method hashCode.
     *
     * @param value
     *            double
     * @return int
     */
    public static int hashCode(double value) {
        long bits = Double.doubleToLongBits(value);

        return (int) (bits ^ (bits >>> 32));
    }

    /**
     * Method hashCode.
     *
     * @param value
     *            Object
     * @return int
     */
    public static int hashCode(Object value) {
        if (value == null) {
            return 0;
        }

        if (value.getClass().isArray()) {
            int order = getClazzEnumOrder(value.getClass());

            switch (order) {
            case 1:
                return hashCode((boolean[]) value);

            case 2:
                return hashCode((char[]) value);

            case 3:
                return hashCode((byte[]) value);

            case 4:
                return hashCode((short[]) value);

            case 5:
                return hashCode((int[]) value);

            case 6:
                return hashCode((long[]) value);

            case 7:
                return hashCode((float[]) value);

            case 8:
                return hashCode((double[]) value);

            default:
                return hashCode((Object[]) value);
            }
        }

        return value.hashCode();
    }

    /**
     * Method deepHashCode.
     *
     * @param value
     *            Object
     * @return int
     */
    public static int deepHashCode(Object value) {
        if (value == null) {
            return 0;
        }

        if (value.getClass().isArray()) {
            int order = getClazzEnumOrder(value.getClass());

            switch (order) {
            case 1:
                return hashCode((boolean[]) value);

            case 2:
                return hashCode((char[]) value);

            case 3:
                return hashCode((byte[]) value);

            case 4:
                return hashCode((short[]) value);

            case 5:
                return hashCode((int[]) value);

            case 6:
                return hashCode((long[]) value);

            case 7:
                return hashCode((float[]) value);

            case 8:
                return hashCode((double[]) value);

            default:
                return deepHashCode((Object[]) value);
            }
        }

        return value.hashCode();
    }

    /**
     * @see Arrays#hashCode(boolean[])
     * @param a
     * @return
     */
    public static int hashCode(boolean[] a) {
        return Arrays.hashCode(a);
    }

    /**
     * @see Arrays#hashCode(char[])
     * @param a
     * @return
     */
    public static int hashCode(char[] a) {
        return Arrays.hashCode(a);
    }

    /**
     * @see Arrays#hashCode(byte[])
     * @param a
     * @return
     */
    public static int hashCode(byte[] a) {
        return Arrays.hashCode(a);
    }

    /**
     * @see Arrays#hashCode(short[])
     * @param a
     * @return
     */
    public static int hashCode(short[] a) {
        return Arrays.hashCode(a);
    }

    /**
     * @see Arrays#hashCode(int[])
     * @param a
     * @return
     */
    public static int hashCode(int[] a) {
        return Arrays.hashCode(a);
    }

    /**
     * @see Arrays#hashCode(long[])
     * @param a
     * @return
     */
    public static int hashCode(long[] a) {
        return Arrays.hashCode(a);
    }

    /**
     * @see Arrays#hashCode(float[])
     * @param a
     * @return
     */
    public static int hashCode(float[] a) {
        return Arrays.hashCode(a);
    }

    /**
     * @see Arrays#hashCode(double[])
     * @param a
     * @return
     */
    public static int hashCode(double[] a) {
        return Arrays.hashCode(a);
    }

    /**
     * @see Arrays#hashCode(Object[])
     * @param a
     * @return
     */
    public static int hashCode(Object[] a) {
        return Arrays.hashCode(a);
    }

    /**
     * @see Arrays#deepHashCode(Object[])
     * @param a
     * @return
     */
    public static int deepHashCode(Object[] a) {
        return Arrays.deepHashCode(a);
    }

    /**
     * Method toString.
     *
     * @param value
     *            boolean
     * @return int
     */
    public static String toString(boolean value) {
        return String.valueOf(value);
    }

    /**
     * Method toString.
     *
     * @param value
     *            char
     * @return int
     */
    public static String toString(char value) {
        return String.valueOf(value);
    }

    /**
     * Method toString.
     *
     * @param value
     *            byte
     * @return int
     */
    public static String toString(byte value) {
        return String.valueOf(value);
    }

    /**
     * Method toString.
     *
     * @param value
     * @return int
     */
    public static String toString(short value) {
        return String.valueOf(value);
    }

    /**
     * Method toString.
     *
     * @param vaue
     *            int
     * @return int
     */
    public static String toString(int value) {
        return String.valueOf(value);
    }

    /**
     * Method toString.
     *
     * @param value
     *            long
     * @return int
     */
    public static String toString(long value) {
        return String.valueOf(value);
    }

    /**
     * Method toString.
     *
     * @param value
     *            float
     * @return int
     */
    public static String toString(float value) {
        return String.valueOf(value);
    }

    /**
     * Method toString.
     *
     * @param value
     *            double
     * @return int
     */
    public static String toString(double value) {
        return String.valueOf(value);
    }

    /**
     * Method toString.
     *
     * @param value
     *            Object
     * @return int
     */
    public static String toString(Object value) {
        if (value == null) {
            return NULL_STRING;
        }

        if (value.getClass().isArray()) {
            int order = getClazzEnumOrder(value.getClass());

            switch (order) {
            case 1:
                return toString((boolean[]) value);

            case 2:
                return toString((char[]) value);

            case 3:
                return toString((byte[]) value);

            case 4:
                return toString((short[]) value);

            case 5:
                return toString((int[]) value);

            case 6:
                return toString((long[]) value);

            case 7:
                return toString((float[]) value);

            case 8:
                return toString((double[]) value);

            default:
                return toString((Object[]) value);
            }
        }

        return value.toString();
    }

    /**
     * Method deepToString.
     *
     * @param value
     *            Object
     * @return int
     */
    public static String deepToString(Object value) {
        if (value == null) {
            return NULL_STRING;
        }

        if (value.getClass().isArray()) {
            int order = getClazzEnumOrder(value.getClass());

            switch (order) {
            case 1:
                return toString((boolean[]) value);

            case 2:
                return toString((char[]) value);

            case 3:
                return toString((byte[]) value);

            case 4:
                return toString((short[]) value);

            case 5:
                return toString((int[]) value);

            case 6:
                return toString((long[]) value);

            case 7:
                return toString((float[]) value);

            case 8:
                return toString((double[]) value);

            default:
                return deepToString((Object[]) value);
            }
        }

        return value.toString();
    }

    /**
     * @see Arrays#toString(boolean[])
     * @param a
     * @return
     */
    public static String toString(boolean[] a) {
        return Arrays.toString(a);
    }

    /**
     * @see Arrays#toString(char[])
     * @param a
     * @return
     */
    public static String toString(char[] a) {
        return Arrays.toString(a);
    }

    /**
     * @see Arrays#toString(byte[])
     * @param a
     * @return
     */
    public static String toString(byte[] a) {
        return Arrays.toString(a);
    }

    /**
     * @see Arrays#toString(short[])
     * @param a
     * @return
     */
    public static String toString(short[] a) {
        return Arrays.toString(a);
    }

    /**
     * @see Arrays#toString(int[])
     * @param a
     * @return
     */
    public static String toString(int[] a) {
        return Arrays.toString(a);
    }

    /**
     * @see Arrays#toString(long[])
     * @param a
     * @return
     */
    public static String toString(long[] a) {
        return Arrays.toString(a);
    }

    /**
     * @see Arrays#toString(float[])
     * @param a
     * @return
     */
    public static String toString(float[] a) {
        return Arrays.toString(a);
    }

    /**
     * @see Arrays#toString(double[])
     * @param a
     * @return
     */
    public static String toString(double[] a) {
        return Arrays.toString(a);
    }

    /**
     * @see Arrays#toString(Object[])
     * @param a
     * @return
     */
    public static String toString(Object[] a) {
        return Arrays.toString(a);
    }

    /**
     * @see Arrays#deepToString(Object[])
     * @param a
     * @return
     */
    public static String deepToString(Object[] a) {
        return Arrays.deepToString(a);
    }

    /**
     * Method newInstance.
     *
     * @param cls
     *            Class<T>
     * @return T
     * @throws AbacusException
     */
    public static <T> T newInstance(Class<T> cls) {
        try {
            return cls.newInstance();
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        } catch (InstantiationException e) {
            throw new RuntimeException(e);
        }
    }

    public static <T> T newProxyInstance(Class<T> interfaceClass, InvocationHandler h) {
        return (T) newProxyInstance(N.asArray(interfaceClass), h);
    }

    /**
     * Refer to {@code java.lang.reflect}
     *
     * @param interfaces
     * @param h
     * @return
     */
    public static Object newProxyInstance(Class<?>[] interfaces, InvocationHandler h) {
        return Proxy.newProxyInstance(N.class.getClassLoader(), interfaces, h);
    }

    /**
     * Method newArray.
     *
     * @param cls
     *            Class<T>
     * @param length
     *            int
     * @return T[]
     */
    public static <T> T newArray(Class<?> cls, int length) {
        return (T) Array.newInstance(cls, length);
    }

    public static <T> ArrayList<T> newArrayList(int initialCapacity) {
        return new ArrayList<T>(initialCapacity);
    }

    public static <T> LinkedList<T> newLinkedList() {
        return new LinkedList<T>();
    }

    public static <T> HashSet<T> newHashSet(int initialCapacity) {
        return new HashSet<T>(initialCapacity);
    }

    public static <T> LinkedHashSet<T> newLinkedHashSet(int initialCapacity) {
        return new LinkedHashSet<T>(initialCapacity);
    }

    public static <K, V> HashMap<K, V> newHashMap(int initialCapacity) {
        return new HashMap<K, V>(initialCapacity);
    }

    public static <K, V> LinkedHashMap<K, V> newLinkedHashMap(int initialCapacity) {
        return new LinkedHashMap<K, V>(initialCapacity);
    }

    public static <K, V> ConcurrentHashMap<K, V> newConcurrentHashMap(int initialCapacity) {
        return new ConcurrentHashMap<K, V>(initialCapacity);
    }

    public static <K, V> IdentityHashMap<K, V> newIdentityHashMap(int initialCapacity) {
        return new IdentityHashMap<K, V>(initialCapacity);
    }

    public static <T> ArrayBlockingQueue<T> newArrayBlockingQueue(int capacity) {
        return new ArrayBlockingQueue<T>(capacity);
    }

    public static <T> LinkedBlockingQueue<T> newLinkedBlockingQueue(int capacity) {
        return new LinkedBlockingQueue<T>(capacity);
    }

    public static <T> ArrayDeque newArrayDeque(int initialCapacity) {
        return new ArrayDeque<T>(initialCapacity);
    }

    public static <T> LinkedBlockingDeque newLinkedBlockingDeque(int capacity) {
        return new LinkedBlockingDeque<T>(capacity);
    }

    public static <T> T newEntity(Class<T> cls, String entityName) {
        try {
            if ((getEnclosingClass(cls) == null) || Modifier.isStatic(cls.getModifiers())) {
                return newInstance(cls);
            } else {
                Object object = getEnclosingClass(cls).newInstance();

                return cls.getConstructor(getEnclosingClass(cls)).newInstance(object);
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static Map<String, Object> asProps(Object... a) {
        Map<String, Object> props = asMap(a);

        return props;
    }

    public static Map<String, Object> asProps(Map<String, Object> m) {
        Map<String, Object> props = asMap(m);

        return props;
    }

    /**
     *
     * @param a
     *            pairs of property name and value or a Java Entity Object what
     *            allows access to properties using getter and setter methods.
     * @return
     */
    public static <K, V> Map<K, V> asMap(Object... a) {
        return newMap(new HashMap<K, V>(), a);
    }

    public static <K, V> Map<K, V> asMap(Map<?extends K, ?extends V> m) {
        Map<K, V> result = new HashMap<K, V>(m);

        if (N.notNullOrEmpty(m)) {
            result.putAll(m);
        }

        return result;
    }

    public static <K, V> LinkedHashMap<K, V> asLinkedHashMap(Object... a) {
        return (LinkedHashMap) newMap(new LinkedHashMap<K, V>(), a);
    }

    public static <K, V> LinkedHashMap<K, V> asLinkedHashMap(Map<?extends K, ?extends V> m) {
        return new LinkedHashMap<K, V>(m);
    }

    public static <K, V> ConcurrentHashMap<K, V> asConcurrentHashMap(Object... a) {
        return (ConcurrentHashMap) newMap(new ConcurrentHashMap<K, V>(), a);
    }

    public static <K, V> ConcurrentHashMap<K, V> asConcurrentHashMap(Map<?extends K, ?extends V> m) {
        return new ConcurrentHashMap<K, V>(m);
    }

    public static <K, V> IdentityHashMap<K, V> asIdentityHashMap(Object... a) {
        return (IdentityHashMap) newMap(new IdentityHashMap<K, V>(), a);
    }

    public static <K, V> IdentityHashMap<K, V> asIdentityHashMap(Map<?extends K, ?extends V> m) {
        return new IdentityHashMap<K, V>(m);
    }

    public static <K, V> SortedMap<K, V> asSortedMap(Object... a) {
        return (SortedMap) newMap(new TreeMap<K, V>(), a);
    }

    public static <K, V> SortedMap<K, V> asSortedMap(Map<?extends K, ?extends V> m) {
        if (m instanceof SortedMap) {
            return new TreeMap<K, V>((SortedMap) m);
        } else {
            return new TreeMap<K, V>(m);
        }
    }

    public static <K, V> SortedMap<K, V> asSortedMap(Comparator<?super K> comparator, Object... a) {
        SortedMap<K, V> sortedMap = new TreeMap<K, V>(comparator);

        return (SortedMap) newMap(sortedMap, a);
    }

    public static <K, V> SortedMap<K, V> asSortedMap(Comparator<?super K> comparator, Map<?extends K, ?extends V> m) {
        SortedMap<K, V> sortedMap = new TreeMap<K, V>(comparator);
        sortedMap.putAll(m);

        return sortedMap;
    }

    private static <K, V> Map<K, V> newMap(Map<K, V> m, Object... a) {
        if (N.isNullOrEmpty(a)) {
            return m;
        }

        if (a.length == 1) {
            if (a[0] instanceof Map) {
                m.putAll((Map) a[0]);

                return m;
            }

            Object entity = a[0];
            Map<String, Method> getterMethodList = N.getPropGetMethodList(a[0].getClass());

            if (getterMethodList.size() == 0) {
                throw new IllegalArgumentException(
                    "The parameter must be the pairs of property name and value, or an entity class with get/set methods.");
            }

            try {
                for (String propName : getterMethodList.keySet()) {
                    m.put((K) propName, (V) getterMethodList.get(propName).invoke(entity));
                }
            } catch (java.lang.IllegalArgumentException e) {
                throw new RuntimeException(e);
            } catch (IllegalAccessException e) {
                throw new RuntimeException(e);
            } catch (InvocationTargetException e) {
                throw new RuntimeException(e);
            }

            return m;
        } else {
            if (0 != (a.length % 2)) {
                throw new IllegalArgumentException(
                    "the parameter must be the pairs of property name and value, or an entity class with get/set methods.");
            }

            for (int i = 0; i < a.length; i++) {
                m.put((K) a[i], (V) a[++i]);
            }

            return m;
        }
    }

    public static <T> List<T> asList(T... a) {
        List<T> list = (a.length > 0) ? new ArrayList<T>(a.length) : new ArrayList();

        for (T e : a) {
            list.add(e);
        }

        return list;
    }

    public static <T> List<T> asList(Collection<T> c) {
        return new ArrayList<T>(c);
    }

    public static <T> LinkedList<T> asLinkedList(T... a) {
        LinkedList<T> list = new LinkedList<T>();

        for (T e : a) {
            list.add(e);
        }

        return list;
    }

    public static <T> LinkedList<T> asLinkedList(Collection<T> c) {
        return new LinkedList<T>(c);
    }

    public static <T> Set<T> asSet(T... a) {
        Set<T> set = new HashSet<T>();

        for (T e : a) {
            set.add(e);
        }

        return set;
    }

    public static <T> Set<T> asSet(Collection<T> c) {
        return new HashSet<T>(c);
    }

    public static <T> LinkedHashSet<T> asLinkedHashSet(T... a) {
        LinkedHashSet<T> set = new LinkedHashSet<T>();

        for (T e : a) {
            set.add(e);
        }

        return set;
    }

    public static <T> LinkedHashSet<T> asLinkedHashSet(Collection<T> c) {
        return new LinkedHashSet<T>(c);
    }

    public static <T> SortedSet<T> asSortedSet(T... a) {
        SortedSet<T> set = new TreeSet<T>();

        for (T e : a) {
            set.add(e);
        }

        return set;
    }

    public static <T> SortedSet<T> asSortedSet(Collection<T> c) {
        if (c instanceof SortedSet) {
            return new TreeSet<T>((SortedSet) c);
        } else {
            return new TreeSet<T>(c);
        }
    }

    public static <T> SortedSet<T> asSortedSet(Comparator<T> comparator, T... a) {
        SortedSet<T> set = new TreeSet<T>(comparator);

        for (T e : a) {
            set.add(e);
        }

        return set;
    }

    public static <T> SortedSet<T> asSortedSet(Comparator<T> comparator, Collection<T> c) {
        SortedSet<T> set = new TreeSet<T>(comparator);
        set.addAll(c);

        return set;
    }

    public static <T> Queue<T> asQueue(T... a) {
        Queue<T> queue = new ArrayDeque<T>();

        for (T e : a) {
            queue.add(e);
        }

        return queue;
    }

    public static <T> Queue<T> asQueue(Collection<T> c) {
        return new ArrayDeque<T>(c);
    }

    public static <T> ArrayBlockingQueue<T> asArrayBlockingQueue(int capacity, T... a) {
        ArrayBlockingQueue<T> queue = new ArrayBlockingQueue<T>(capacity);

        for (T e : a) {
            queue.add(e);
        }

        return queue;
    }

    public static <T> ArrayBlockingQueue<T> asArrayBlockingQueue(int capacity, Collection<T> c) {
        return new ArrayBlockingQueue<T>(capacity, false, c);
    }

    public static <T> LinkedBlockingQueue<T> asLinkedBlockingQueue(T... a) {
        LinkedBlockingQueue<T> queue = new LinkedBlockingQueue<T>();

        for (T e : a) {
            queue.add(e);
        }

        return queue;
    }

    public static <T> LinkedBlockingQueue<T> asLinkedBlockingQueue(Collection<T> c) {
        return new LinkedBlockingQueue<T>(c);
    }

    public static <T> ConcurrentLinkedQueue<T> asConcurrentLinkedQueue(T... a) {
        ConcurrentLinkedQueue<T> queue = new ConcurrentLinkedQueue<T>();

        for (T e : a) {
            queue.add(e);
        }

        return queue;
    }

    public static <T> ConcurrentLinkedQueue<T> asConcurrentLinkedQueue(Collection<T> c) {
        return new ConcurrentLinkedQueue<T>(c);
    }

    public static <T extends Delayed> DelayQueue<T> asDelayQueue(T... a) {
        DelayQueue<T> queue = new DelayQueue<T>();

        for (T e : a) {
            queue.add(e);
        }

        return queue;
    }

    public static <T extends Delayed> DelayQueue<T> asDelayQueue(Collection<T> c) {
        return new DelayQueue<T>(c);
    }

    public static <T> Deque<T> asDeque(T... a) {
        Deque<T> deque = new ArrayDeque<T>();

        for (T e : a) {
            deque.add(e);
        }

        return deque;
    }

    public static <T> Deque<T> asDeque(Collection<T> c) {
        return new ArrayDeque<T>(c);
    }

    public static <T> LinkedBlockingDeque<T> asLinkedBlockingDeque(T... a) {
        LinkedBlockingDeque<T> deque = new LinkedBlockingDeque<T>();

        for (T e : a) {
            deque.add(e);
        }

        return deque;
    }

    public static <T> LinkedBlockingDeque<T> asLinkedBlockingDeque(Collection<T> c) {
        return new LinkedBlockingDeque<T>(c);
    }

    public static <T> T[] asArray(T... a) {
        return a;
    }

    @SuppressWarnings("unchecked")
    public static <T> T[] asArray(Collection<T> c) {
        if ((c != null) && (c.size() > 0)) {
            return c.toArray((T[]) Array.newInstance(c.iterator().next().getClass(), c.size()));
        } else {
            return (T[]) new Object[0];
        }
    }

    private static <T> boolean isArray(Class<T> cls) {
        return cls.isArray();
    }

    /**
     * Returns a <code>Boolean</code> with a value represented by the specified
     * string. The <code>Boolean</code> returned represents a true value if the
     * string argument is not <code>null</code> and is equal, ignoring case, to
     * the string {@code "true"}.
     *
     * @param st
     *            a string.
     * @return the <code>Boolean</code> value represented by the string.
     */
    public static boolean asBoolean(String st) {
        return Boolean.valueOf(st);
    }

    /**
     * Returns the value by calling {@code Byte.valueOf(String)} if {@code st}
     * is not {@code null}, otherwise, the default value 0 for {@code byte} is
     * returned.
     *
     * @param st
     * @return
     */
    public static byte asByte(String st) {
        return N.isNullOrEmpty(st) ? 0 : Byte.parseByte(st);
    }

    /**
     * Returns the value by calling {@code Short.valueOf(String)} if {@code st}
     * is not {@code null}, otherwise, the default value 0 for {@code byte} is
     * returned.
     *
     * @param st
     * @return
     */
    public static int asShort(String st) {
        return N.isNullOrEmpty(st) ? 0 : Short.parseShort(st);
    }

    /**
     * Returns the value by calling {@code Integer.valueOf(String)} if
     * {@code st} is not {@code null}, otherwise, the default value 0 for
     * {@code byte} is returned.
     *
     * @param st
     * @return
     */
    public static int asInt(String st) {
        return N.isNullOrEmpty(st) ? 0 : Integer.parseInt(st);
    }

    /**
     * Returns the value by calling {@code Long.valueOf(String)} if {@code st}
     * is not {@code null}, otherwise, the default value 0 for {@code byte} is
     * returned.
     *
     * @param st
     * @return
     */
    public static long asLong(String st) {
        return N.isNullOrEmpty(st) ? 0 : Long.parseLong(st);
    }

    /**
     * Returns the value by calling {@code Float.valueOf(String)} if {@code st}
     * is not {@code null}, otherwise, the default value 0f for {@code byte} is
     * returned.
     *
     * @param st
     * @return
     */
    public static float asFloat(String st) {
        return N.isNullOrEmpty(st) ? 0 : Float.parseFloat(st);
    }

    /**
     * Returns the value by calling {@code Double.valueOf(String)} if {@code st}
     * is not {@code null}, otherwise, the default value 0d for {@code byte} is
     * returned.
     *
     * @param st
     * @return
     */
    public static double asDouble(String st) {
        return N.isNullOrEmpty(st) ? 0 : Double.parseDouble(st);
    }

    public static java.util.Date asJUDate(Calendar c) {
        return (c == null) ? null : asJUDate(c.getTimeInMillis());
    }

    public static java.util.Date asJUDate(java.util.Date date) {
        return (date == null) ? null : asJUDate(date.getTime());
    }

    public static java.util.Date asJUDate(long timeInMillis) {
        return (timeInMillis == 0) ? null : new java.util.Date(timeInMillis);
    }

    public static java.util.Date asJUDate(String date) {
        return asJUDate(date, null);
    }

    public static java.util.Date asJUDate(String date, String format) {
        return asJUDate(date, format, null);
    }

    /**
     * Converts the <code>String</code> date in the specified {@code format} to
     * a new instance of Date.
     *
     * @param date
     * @param format
     * @throws IllegalArgumentException
     *             if the date given can't be parsed with specified format.
     */
    public static java.util.Date asJUDate(String date, String format, TimeZone timeZone) {
        if (N.isNullOrEmpty(date)) {
            return null;
        }

        format = checkDateFormat(date, format);
        timeZone = checkTimeZone(format, timeZone);

        long timeInMillis = fastDateParse(date, format, timeZone);

        if (timeInMillis > 0) {
            return new Date(timeInMillis);
        }

        SimpleDateFormat sdf = getSDF(format, timeZone);

        try {
            return sdf.parse(date);
        } catch (ParseException e) {
            throw new RuntimeException(e);
        } finally {
            SDF_POOL.get(format).add(sdf);
        }
    }

    private static long parse(String date, String format, TimeZone timeZone) {
        format = checkDateFormat(date, format);
        timeZone = checkTimeZone(format, timeZone);

        long timeInMillis = fastDateParse(date, format, timeZone);

        if (timeInMillis > 0) {
            return timeInMillis;
        }

        SimpleDateFormat sdf = getSDF(format, timeZone);

        try {
            return sdf.parse(date).getTime();
        } catch (ParseException e) {
            throw new RuntimeException(e);
        } finally {
            SDF_POOL.get(format).add(sdf);
        }
    }

    private static String checkDateFormat(String st, String format) {
        if (format == null) {
            format = (st.length() == 20) ? DEFAULT_DATETIME_FORMAT
                                         : ((st.length() == 24) ? DEFAULT_TIMESTAMP_FORMAT
                                                                : (((st.length() == 19) ? LOCAL_DATETIME_FORMAT : null)));
        }

        if (format == null) {
            throw new RuntimeException("No valid date format found for: " + st);
        }

        return format;
    }

    private static TimeZone checkTimeZone(String format, TimeZone timeZone) {
        if (timeZone == null) {
            timeZone = ((format.charAt(format.length() - 2) == 'Z') && format.endsWith("'Z'")) ? UTC_TIME_ZONE
                                                                                               : LOCAL_TIME_ZONE;
        }

        return timeZone;
    }

    private static final long fastDateParse(String st, String format, TimeZone timeZone) {
        if (!(format.equals(DEFAULT_DATETIME_FORMAT) || format.equals(DEFAULT_TIMESTAMP_FORMAT)
                || format.equals(LOCAL_DATETIME_FORMAT))) {
            return 0;
        }

        if (!((st.length() == 20) || (st.length() == 24) || (st.length() == 19))) {
            return 0;
        }

        //        if (!((st.charAt(4) == '-') && (st.charAt(7) == '-') && (st.charAt(10) == 'T') && (st.charAt(13) == ':')
        //                && (st.charAt(16) == ':') && (st.charAt(st.length() - 1) == 'Z'))) {
        //            return 0;
        //        }

        //        if (!((st.charAt(4) == '-') && (st.charAt(7) == '-') && (st.charAt(10) == 'T') && (st.charAt(13) == ':')
        //                && (st.charAt(16) == ':') && (st.charAt(st.length() - 1) == 'Z'))) {
        //            return 0;
        //        }
        //        
        //        int year = Integer.valueOf(st.substring(0, 4));
        //        int month = Integer.valueOf(st.substring(5, 7)) - 1;
        //        int date = Integer.valueOf(st.substring(8, 10));
        //        int hourOfDay = Integer.valueOf(st.substring(11, 13));
        //        int minute = Integer.valueOf(st.substring(14, 16));
        //        int second = Integer.valueOf(st.substring(17, 19));
        //        int milliSecond = (st.length() == 24)
        //            ? Integer.valueOf(st.substring(20, 23)) : 0;
        //
        int year = digit(st, 0, 4);
        int month = digit(st, 5, 7) - 1;
        int date = digit(st, 8, 10);
        int hourOfDay = digit(st, 11, 13);
        int minute = digit(st, 14, 16);
        int second = digit(st, 17, 19);
        int milliSecond = (st.length() == 24) ? digit(st, 20, 23) : 0;
        Calendar c = null;
        Queue<Calendar> timeZoneCalendarQueue = null;

        if (timeZone == UTC_TIME_ZONE) {
            c = utcCalendarPool.poll();
        } else {
            timeZoneCalendarQueue = calendarPool.get(timeZone);

            if (timeZoneCalendarQueue == null) {
                timeZoneCalendarQueue = N.newArrayBlockingQueue(SMALL_POOL_SIZE);
                calendarPool.put(timeZone, timeZoneCalendarQueue);
            } else {
                c = timeZoneCalendarQueue.poll();
            }
        }

        if (c == null) {
            c = Calendar.getInstance(timeZone);
        }

        c.set(year, month, date, hourOfDay, minute, second);
        c.set(Calendar.MILLISECOND, milliSecond);

        long timeInMillis = c.getTimeInMillis();

        if (timeZone == UTC_TIME_ZONE) {
            utcCalendarPool.add(c);
        } else {
            timeZoneCalendarQueue.add(c);
        }

        return timeInMillis;
    }

    private static int digit(String st, int from, int to) {
        int result = 0;

        while (from < to) {
            result = (result * 10) + (st.charAt(from++) - 48);
        }

        return result;
    }

    public static Timestamp asTimestamp(Calendar c) {
        return (c == null) ? null : asTimestamp(c.getTimeInMillis());
    }

    public static Timestamp asTimestamp(java.util.Date date) {
        return (date == null) ? null : asTimestamp(date.getTime());
    }

    public static Timestamp asTimestamp(long timeInMillis) {
        return (timeInMillis == 0) ? null : new Timestamp(timeInMillis);
    }

    public static Timestamp asTimestamp(String date) {
        return asTimestamp(date, null);
    }

    public static Timestamp asTimestamp(String date, String format) {
        return asTimestamp(date, format, null);
    }

    public static Timestamp asTimestamp(String date, String format, TimeZone timeZone) {
        if (N.isNullOrEmpty(date)) {
            return null;
        }

        return asTimestamp(parse(date, format, timeZone));
    }

    public static Date asDate(Calendar c) {
        return (c == null) ? null : asDate(c.getTimeInMillis());
    }

    public static Date asDate(java.util.Date date) {
        return (date == null) ? null : asDate(date.getTime());
    }

    public static Date asDate(long timeInMillis) {
        return (timeInMillis == 0) ? null : new Date(timeInMillis);
    }

    public static Date asDate(String date) {
        return asDate(date, null);
    }

    public static Date asDate(String date, String format) {
        return asDate(date, format, null);
    }

    public static Date asDate(String date, String format, TimeZone timeZone) {
        if (N.isNullOrEmpty(date)) {
            return null;
        }

        return asDate(parse(date, format, timeZone));
    }

    public static Time asTime(Calendar c) {
        return (c == null) ? null : asTime(c.getTimeInMillis());
    }

    public static Time asTime(java.util.Date date) {
        return (date == null) ? null : asTime(date.getTime());
    }

    public static Time asTime(long timeInMillis) {
        return (timeInMillis == 0) ? null : new Time(timeInMillis);
    }

    public static Time asTime(String date) {
        return asTime(date, null);
    }

    public static Time asTime(String date, String format) {
        return asTime(date, format, null);
    }

    public static Time asTime(String date, String format, TimeZone timeZone) {
        if (N.isNullOrEmpty(date)) {
            return null;
        }

        return asTime(parse(date, format, timeZone));
    }

    public static Calendar asCalendar(Calendar c) {
        return (c == null) ? null : asCalendar(c.getTimeInMillis());
    }

    public static Calendar asCalendar(java.util.Date date) {
        return (date == null) ? null : asCalendar(date.getTime());
    }

    public static Calendar asCalendar(long timeInMillis) {
        if (timeInMillis == 0) {
            return null;
        }

        Calendar c = Calendar.getInstance();
        c.setTimeInMillis(timeInMillis);

        return c;
    }

    public static Calendar asCalendar(String calendar) {
        return asCalendar(calendar, null);
    }

    public static Calendar asCalendar(String calendar, String format) {
        return asCalendar(calendar, format, null);
    }

    public static Calendar asCalendar(String calendar, String format, TimeZone timeZone) {
        if (N.isNullOrEmpty(calendar)) {
            return null;
        }

        return asCalendar(parse(calendar, format, timeZone));
    }

    public static GregorianCalendar asGregorianCalendar(Calendar c) {
        return (c == null) ? null : asGregorianCalendar(c.getTimeInMillis());
    }

    public static GregorianCalendar asGregorianCalendar(java.util.Date date) {
        return (date == null) ? null : asGregorianCalendar(date.getTime());
    }

    public static GregorianCalendar asGregorianCalendar(long timeInMillis) {
        if (timeInMillis == 0) {
            return null;
        }

        GregorianCalendar c = new GregorianCalendar();
        c.setTimeInMillis(timeInMillis);

        return c;
    }

    public static GregorianCalendar asGregorianCalendar(String calendar) {
        return asGregorianCalendar(calendar, null);
    }

    public static GregorianCalendar asGregorianCalendar(String calendar, String format) {
        return asGregorianCalendar(calendar, format, null);
    }

    public static GregorianCalendar asGregorianCalendar(String calendar, String format, TimeZone timeZone) {
        if (N.isNullOrEmpty(calendar)) {
            return null;
        }

        return asGregorianCalendar(parse(calendar, format, timeZone));
    }

    public static XMLGregorianCalendar asXMLGregorianCalendar(Calendar c) {
        return (c == null) ? null : asXMLGregorianCalendar(c.getTimeInMillis());
    }

    public static XMLGregorianCalendar asXMLGregorianCalendar(java.util.Date date) {
        return (date == null) ? null : asXMLGregorianCalendar(date.getTime());
    }

    public static XMLGregorianCalendar asXMLGregorianCalendar(long timeInMillis) {
        if (timeInMillis == 0) {
            return null;
        }

        XMLGregorianCalendar c = dataTypeFactory.newXMLGregorianCalendar(N.asGregorianCalendar(timeInMillis));

        return c;
    }

    public static XMLGregorianCalendar asXMLGregorianCalendar(String calendar) {
        return asXMLGregorianCalendar(calendar, null);
    }

    public static XMLGregorianCalendar asXMLGregorianCalendar(String calendar, String format) {
        return asXMLGregorianCalendar(calendar, format, null);
    }

    public static XMLGregorianCalendar asXMLGregorianCalendar(String calendar, String format, TimeZone timeZone) {
        if (N.isNullOrEmpty(calendar)) {
            return null;
        }

        return asXMLGregorianCalendar(parse(calendar, format, timeZone));
    }

    /**
     * @see System#currentTimeMillis()
     * @return
     */
    public static long currentMillis() {
        return System.currentTimeMillis();
    }

    /**
     * A new instance of <code>java.sql.Time</code> returned is based on the
     * current time in the default time zone with the default locale.
     *
     * @return
     */
    public static Time currentTime() {
        return new Time(System.currentTimeMillis());
    }

    /**
     * A new instance of <code>java.sql.Date</code> returned is based on the
     * current time in the default time zone with the default locale.
     *
     * @return
     */
    public static Date currentDate() {
        return new Date(System.currentTimeMillis());
    }

    /**
     * A new instance of <code>java.sql.Timestamp</code> returned is based on
     * the current time in the default time zone with the default locale.
     *
     * @return
     */
    public static Timestamp currentTimestamp() {
        return new Timestamp(System.currentTimeMillis());
    }

    /**
     * A new instance of <code>java.util.Date</code> returned is based on the
     * current time in the default time zone with the default locale.
     *
     * @return
     */
    public static java.util.Date currentJUDate() {
        return new java.util.Date();
    }

    /**
     * A new instance of <code>java.util.Calendar</code> returned is based on
     * the current time in the default time zone with the default locale.
     *
     * @return a Calendar.
     */
    public static Calendar currentCalendar() {
        return Calendar.getInstance();
    }

    public static GregorianCalendar currentGregorianCalendar() {
        return new GregorianCalendar();
    }

    public static XMLGregorianCalendar currentXMLGregorianCalendar() {
        return dataTypeFactory.newXMLGregorianCalendar(currentGregorianCalendar());
    }

    private static SimpleDateFormat getSDF(String format, TimeZone timeZone) {
        Queue<SimpleDateFormat> queue = SDF_POOL.get(format);

        if (queue == null) {
            queue = N.newArrayBlockingQueue(SMALL_POOL_SIZE);
            SDF_POOL.put(format, queue);
        }

        SimpleDateFormat sdf = queue.poll();

        if (sdf == null) {
            sdf = new SimpleDateFormat(format);
        }

        sdf.setTimeZone(timeZone);

        return sdf;
    }

    /**
     * Adds or subtracts the specified amount of time to the given time unit,
     * based on the calendar's rules. For example, to subtract 5 days from the
     * current time of the calendar, you can achieve it by calling:
     * <p>
     * <code>add(TimeUnit.DAY, -5)</code>.
     *
     * @param date
     * @param amount
     * @param unit
     * @return the input instance of Date is returned.
     */
    public static <T extends java.util.Date> T roll(T date, long amount, TimeUnit unit) {
        date.setTime(date.getTime() + unit.toMillis(amount));

        return date;
    }

    /**
     * Adds or subtracts the specified amount of time to the given calendar
     * unit, based on the calendar's rules. For example, to subtract 5 days from
     * the current time of the calendar, you can achieve it by calling:
     * <p>
     * <code>add(CalendarUnit.DAY, -5)</code>.
     *
     * @param date
     * @param amount
     * @param unit
     * @return the input instance of Date is returned.
     */
    public static <T extends java.util.Date> T roll(T date, int amount, CalendarUnit unit) {
        switch (unit) {
        case MONTH:
        case YEAR:

            Calendar c = N.asCalendar(date);
            c.add(unit.intValue(), amount);
            date.setTime(c.getTimeInMillis());

            break;

        default:
            date.setTime(date.getTime() + unit.toMillis(amount));
        }

        return date;
    }

    /**
     * Adds or subtracts the specified amount of time to the given time unit,
     * based on the calendar's rules. For example, to subtract 5 days from the
     * current time of the calendar, you can achieve it by calling:
     * <p>
     * <code>add(TimeUnit.DAY, -5)</code>.
     *
     * @param c
     * @param amount
     * @param unit
     * @return the input instance of Calendar is returned.
     */
    public static void roll(Calendar c, long amount, TimeUnit unit) {
        c.setTimeInMillis(c.getTimeInMillis() + unit.toMillis(amount));
    }

    /**
     * Adds or subtracts the specified amount of time to the given calendar
     * unit, based on the calendar's rules. For example, to subtract 5 days from
     * the current time of the calendar, you can achieve it by calling:
     * <p>
     * <code>add(CalendarUnit.DAY, -5)</code>.
     *
     * @param c
     * @param amount
     * @param unit
     * @return the input instance of Calendar is returned.
     */
    public static Calendar roll(Calendar c, int amount, CalendarUnit unit) {
        c.add(unit.intValue(), amount);

        return c;
    }

    /**
     * @see Base64#encode(byte[])
     * @param binaryData
     * @return
     */
    public static String base64Encode(byte[] binaryData) {
        return Base64.encode(binaryData);
    }

    /**
     * @see Base64#decode(String);
     * @param encoded
     * @return
     */
    public static byte[] base64Decode(String encoded) {
        return Base64.decode(encoded);
    }

    /**
     * Returns an UUID
     *
     * @see UUID#randomUUID().
     * @return
     */
    public static String uuid() {
        SecureRandom ng = numberGenerator;

        byte[] randomBytes = new byte[16];
        ng.nextBytes(randomBytes);
        randomBytes[6] &= 0x0f; /* clear version */
        randomBytes[6] |= 0x40; /* set to version 4 */
        randomBytes[8] &= 0x3f; /* clear variant */
        randomBytes[8] |= 0x80; /* set to IETF variant */

        long mostSigBits = 0;

        for (int i = 0; i < 8; i++) {
            mostSigBits = (mostSigBits << 8) | (randomBytes[i] & 0xff);
        }

        long leastSigBits = 0;

        for (int i = 8; i < 16; i++) {
            leastSigBits = (leastSigBits << 8) | (randomBytes[i] & 0xff);
        }

        return (digits(mostSigBits >> 32, 8) + digits(mostSigBits >> 16, 4) + digits(mostSigBits, 4)
        + digits(leastSigBits >> 48, 4) + digits(leastSigBits, 12));
    }

    /** Returns val represented by the specified number of hex digits. */
    private static String digits(long val, int digits) {
        long hi = 1L << (digits * 4);

        return Long.toHexString(hi | (val & (hi - 1))).substring(1);
    }

    /**
     * Returns Base64 encoded UID that is unique over time with respect to the
     * host that it was generated on.
     *
     * @see UID#UID()
     * @return
     */
    public static String vuid() {
        return Base64.encode(new UID().toString().getBytes());
    }

    /**
     * Returns a new String with upper case first char if {@code str} is not
     * null or empty, otherwise the original String is returned.
     *
     * @param str
     * @return
     */
    public static String upCaseFirstChar(String str) {
        if (isNullOrEmpty(str)) {
            return str;
        } else if (str.length() == 1) {
            return toString(Character.toUpperCase(str.charAt(0)));
        } else {
            return Character.toUpperCase(str.charAt(0)) + str.substring(1);
        }
    }

    /**
     * Returns a new String with lower case first char if {@code str} is not
     * null or empty, otherwise the original String is returned.
     *
     * @param str
     * @return
     */
    public static String lowCaseFirstChar(String str) {
        if (isNullOrEmpty(str)) {
            return str;
        } else if (str.length() == 1) {
            return toString(Character.toLowerCase(str.charAt(0)));
        } else {
            return Character.toLowerCase(str.charAt(0)) + str.substring(1);
        }
    }

    public static <T> Future<T> asynExecute(final Method method, final Object instance, final Object... args) {
        FutureTask<T> future = new FutureTask<T>(new Callable<T>() {
                    public T call() throws Exception {
                        return (T) method.invoke(instance, args);
                    }
                });

        executorService.execute(future);

        return future;
    }

    /**
     *
     * @param command
     */
    public static void asynExecute(Runnable command) {
        executorService.execute(command);
    }

    public static void println(Object obj) {
        System.out.println(N.toString(obj));
    }

    public static void sleep(long millis) {
        try {
            Thread.sleep(millis);
        } catch (InterruptedException e) {
            logger.warning("failed to sleep: " + millis + " due to: " + e.getMessage());
        }
    }

    static Class<?>[] getTypeArgumentsBySetMethod(Method propSetMethod) {
        if (propSetMethod == null) {
            return null;
        }

        Class<?>[] typeParameterClasses = methodTypeArgumentsPool.get(propSetMethod);

        if (typeParameterClasses == null) {
            java.lang.reflect.Type genericParameterType = propSetMethod.getGenericParameterTypes()[0];

            if (genericParameterType instanceof ParameterizedType) {
                ParameterizedType aType = (ParameterizedType) genericParameterType;
                java.lang.reflect.Type[] parameterArgTypes = aType.getActualTypeArguments();
                typeParameterClasses = new Class[parameterArgTypes.length];

                for (int i = 0; i < parameterArgTypes.length; i++) {
                    typeParameterClasses[i] = (Class<?>) parameterArgTypes[i];
                }
            } else {
                typeParameterClasses = DUMMY_CLASSES;
            }

            methodTypeArgumentsPool.put(propSetMethod, typeParameterClasses);
        }

        return (typeParameterClasses == DUMMY_CLASSES) ? null : typeParameterClasses;
    }

    private static <T> void checkEntityClass(Class<T> cls) {
        if (N.getPropGetMethodList(cls).size() == 0) {
            throw new IllegalArgumentException("No property getXXX/setXXX method is found in the specified class: "
                + N.getClassCanonicalName(cls));
        }
    }

    public static String list2String(List<?> list) {
        return collection2String(list, ", ", false);
    }

    public static String list2String(List<?> list, String separator) {
        return collection2String(list, separator, false);
    }

    public static String list2String(List<?> list, String separator, boolean trim) {
        return collection2String(list, separator, trim);
    }

    public static String set2String(Set<?> set) {
        return collection2String(set, ", ", false);
    }

    public static String set2String(Set<?> set, String separator) {
        return collection2String(set, separator, false);
    }

    public static String set2String(Set<?> set, String separator, boolean trim) {
        return collection2String(set, separator, trim);
    }

    public static String queue2String(Queue<?> queue) {
        return collection2String(queue, ", ", false);
    }

    public static String queue2String(Queue<?> queue, String separator) {
        return collection2String(queue, separator, false);
    }

    public static String queue2String(Queue<?> queue, String separator, boolean trim) {
        return collection2String(queue, separator, trim);
    }

    private static String collection2String(Collection<?> c, String separator, boolean trim) {
        if (c == null) {
            return null;
        }

        Iterator<?> it = c.iterator();

        if (trim) {
            if (c.size() == 1) {
                return toString(it.next()).trim();
            } else if (c.size() == 2) {
                return toString(it.next()).trim() + separator + toString(it.next()).trim();
            } else if (c.size() == 3) {
                return toString(it.next()).trim() + separator + toString(it.next()).trim() + separator
                + toString(it.next()).trim();
            } else {
                StringBuilder sb = ObjectFactory.createStringBuilder();

                for (int i = 0, size = c.size(); i < size; i++) {
                    if (i > 0) {
                        sb.append(separator);
                    }

                    sb.append(toString(it.next()).trim());
                }

                String st = sb.toString();
                ObjectFactory.recycle(sb);

                return st;
            }
        } else {
            if (c.size() == 1) {
                return toString(it.next());
            } else if (c.size() == 2) {
                return toString(it.next()) + separator + toString(it.next());
            } else if (c.size() == 3) {
                return toString(it.next()) + separator + toString(it.next()) + separator + toString(it.next());
            } else {
                StringBuilder sb = ObjectFactory.createStringBuilder();

                for (int i = 0, size = c.size(); i < size; i++) {
                    if (i > 0) {
                        sb.append(separator);
                    }

                    sb.append(toString(it.next()));
                }

                String st = sb.toString();
                ObjectFactory.recycle(sb);

                return st;
            }
        }
    }

    public static String array2String(boolean[] a, String separator) {
        if (a.length == 1) {
            return String.valueOf(a[0]);
        } else if (a.length == 2) {
            return a[0] + separator + a[1];
        } else if (a.length == 3) {
            return a[0] + separator + a[1] + separator + a[2];
        } else {
            StringBuilder sb = new StringBuilder();

            for (int i = 0, length = a.length; i < length; i++) {
                if (i > 0) {
                    sb.append(separator);
                }

                sb.append(a[i]);
            }

            return sb.toString();
        }
    }

    public static String array2String(char[] a, String separator) {
        if (a.length == 1) {
            return String.valueOf(a[0]);
        } else if (a.length == 2) {
            return a[0] + separator + a[1];
        } else if (a.length == 3) {
            return a[0] + separator + a[1] + separator + a[2];
        } else {
            StringBuilder sb = new StringBuilder();

            for (int i = 0, length = a.length; i < length; i++) {
                if (i > 0) {
                    sb.append(separator);
                }

                sb.append(a[i]);
            }

            return sb.toString();
        }
    }

    public static String array2String(byte[] a, String separator) {
        if (a.length == 1) {
            return String.valueOf(a[0]);
        } else if (a.length == 2) {
            return a[0] + separator + a[1];
        } else if (a.length == 3) {
            return a[0] + separator + a[1] + separator + a[2];
        } else {
            StringBuilder sb = new StringBuilder();

            for (int i = 0, length = a.length; i < length; i++) {
                if (i > 0) {
                    sb.append(separator);
                }

                sb.append(a[i]);
            }

            return sb.toString();
        }
    }

    public static String array2String(short[] a, String separator) {
        if (a.length == 1) {
            return String.valueOf(a[0]);
        } else if (a.length == 2) {
            return a[0] + separator + a[1];
        } else if (a.length == 3) {
            return a[0] + separator + a[1] + separator + a[2];
        } else {
            StringBuilder sb = new StringBuilder();

            for (int i = 0, length = a.length; i < length; i++) {
                if (i > 0) {
                    sb.append(separator);
                }

                sb.append(a[i]);
            }

            return sb.toString();
        }
    }

    public static String array2String(int[] a, String separator) {
        if (a.length == 1) {
            return String.valueOf(a[0]);
        } else if (a.length == 2) {
            return a[0] + separator + a[1];
        } else if (a.length == 3) {
            return a[0] + separator + a[1] + separator + a[2];
        } else {
            StringBuilder sb = new StringBuilder();

            for (int i = 0, length = a.length; i < length; i++) {
                if (i > 0) {
                    sb.append(separator);
                }

                sb.append(a[i]);
            }

            return sb.toString();
        }
    }

    public static String array2String(long[] a, String separator) {
        if (a.length == 1) {
            return String.valueOf(a[0]);
        } else if (a.length == 2) {
            return a[0] + separator + a[1];
        } else if (a.length == 3) {
            return a[0] + separator + a[1] + separator + a[2];
        } else {
            StringBuilder sb = new StringBuilder();

            for (int i = 0, length = a.length; i < length; i++) {
                if (i > 0) {
                    sb.append(separator);
                }

                sb.append(a[i]);
            }

            return sb.toString();
        }
    }

    public static String array2String(float[] a, String separator) {
        if (a.length == 1) {
            return String.valueOf(a[0]);
        } else if (a.length == 2) {
            return a[0] + separator + a[1];
        } else if (a.length == 3) {
            return a[0] + separator + a[1] + separator + a[2];
        } else {
            StringBuilder sb = new StringBuilder();

            for (int i = 0, length = a.length; i < length; i++) {
                if (i > 0) {
                    sb.append(separator);
                }

                sb.append(a[i]);
            }

            return sb.toString();
        }
    }

    public static String array2String(double[] a, String separator) {
        if (a.length == 1) {
            return String.valueOf(a[0]);
        } else if (a.length == 2) {
            return a[0] + separator + a[1];
        } else if (a.length == 3) {
            return a[0] + separator + a[1] + separator + a[2];
        } else {
            StringBuilder sb = new StringBuilder();

            for (int i = 0, length = a.length; i < length; i++) {
                if (i > 0) {
                    sb.append(separator);
                }

                sb.append(a[i]);
            }

            return sb.toString();
        }
    }

    public static String array2String(Object a) {
        return array2String(a, ", ", false);
    }

    public static String array2String(Object a, String separator) {
        return array2String(a, separator, false);
    }

    public static String array2String(Object a, String separator, boolean trim) {
        if (a == null) {
            return null;
        }

        if (!a.getClass().isArray()) {
            throw new IllegalArgumentException(a.getClass() + " is not an array type");
        }

        int order = getClazzEnumOrder(a.getClass());

        switch (order) {
        case 1:

            boolean[] booleans = (boolean[]) a;

            return array2String(booleans, separator);

        case 2:

            char[] chars = (char[]) a;

            return array2String(chars, separator);

        case 3:

            byte[] bytes = (byte[]) a;

            return array2String(bytes, separator);

        case 4:

            short[] shorts = (short[]) a;

            return array2String(shorts, separator);

        case 5:

            int[] ints = (int[]) a;

            return array2String(ints, separator);

        case 6:

            long[] longs = (long[]) a;

            return array2String(longs, separator);

        case 7:

            float[] floats = (float[]) a;

            return array2String(floats, separator);

        case 8:

            double[] doubles = (double[]) a;

            return array2String(doubles, separator);

        default:

            Object[] objects = (Object[]) a;

            if (trim) {
                if (objects.length == 1) {
                    return toString(objects[0]).trim();
                } else if (objects.length == 2) {
                    return toString(objects[0]).trim() + separator + toString(objects[1]).trim();
                } else if (objects.length == 3) {
                    return toString(objects[0]).trim() + separator + toString(objects[1]).trim() + separator
                    + toString(objects[2]).trim();
                } else {
                    StringBuilder sb = ObjectFactory.createStringBuilder();

                    for (int i = 0; i < objects.length; i++) {
                        if (i > 0) {
                            sb.append(separator);
                        }

                        sb.append(toString(objects[i]).trim());
                    }

                    String st = sb.toString();
                    ObjectFactory.recycle(sb);

                    return st;
                }
            } else {
                if (objects.length == 1) {
                    return toString(objects[0]);
                } else if (objects.length == 2) {
                    return toString(objects[0]) + separator + toString(objects[1]);
                } else if (objects.length == 3) {
                    return toString(objects[0]) + separator + toString(objects[1]) + separator + toString(objects[2]);
                } else {
                    StringBuilder sb = ObjectFactory.createStringBuilder();

                    for (int i = 0; i < objects.length; i++) {
                        if (i > 0) {
                            sb.append(separator);
                        }

                        sb.append(toString(objects[i]));
                    }

                    String st = sb.toString();
                    ObjectFactory.recycle(sb);

                    return st;
                }
            }
        }
    }

    public static String format(java.util.Date date) {
        return format(date, null, null);
    }

    public static String format(java.util.Date date, String format) {
        return format(date, format, null);
    }

    public static String format(java.util.Date date, String format, TimeZone timeZone) {
        return formatDate(date, format, timeZone, null);
    }

    public static void format(java.util.Date date, String format, TimeZone timeZone, Writer writer) {
        formatDate(date, format, timeZone, writer);
    }

    private static String formatDate(java.util.Date date, String format, TimeZone timeZone, Writer writer) {
        boolean isTimestamp = date instanceof Timestamp;

        if ((format == null) && (timeZone == null)) {
            if (writer == null) {
                BufferedWriter cbuff = ObjectFactory.createBufferedWriter();
                fastDateFormat(date.getTime(), isTimestamp, cbuff);

                String st = cbuff.toString();
                ObjectFactory.recycle(cbuff);

                return st;
            } else {
                fastDateFormat(date.getTime(), isTimestamp, writer);

                return null;
            }
        }

        if (format == null) {
            format = isTimestamp ? DEFAULT_TIMESTAMP_FORMAT : DEFAULT_DATETIME_FORMAT;
        }

        timeZone = checkTimeZone(format, timeZone);

        SimpleDateFormat sdf = getSDF(format, timeZone);

        String st = sdf.format(date);

        if (writer == null) {
            SDF_POOL.get(format).add(sdf);

            return st;
        } else {
            try {
                writer.write(st);
            } catch (IOException e) {
                throw new RuntimeException();
            } finally {
                SDF_POOL.get(format).add(sdf);
            }

            return null;
        }
    }

    public static String format(Calendar c) {
        return format(c, null, null);
    }

    public static String format(Calendar c, String format) {
        return format(c, format, null);
    }

    public static String format(Calendar c, String format, TimeZone timeZone) {
        if ((format == null) && (timeZone == null)) {
            BufferedWriter cbuff = ObjectFactory.createBufferedWriter();
            fastDateFormat(c.getTimeInMillis(), false, cbuff);

            String st = cbuff.toString();
            ObjectFactory.recycle(cbuff);

            return st;
        }

        return format(N.asJUDate(c), format, timeZone);
    }

    public static void format(Calendar c, String format, TimeZone timeZone, Writer writer) {
        if ((format == null) && (timeZone == null)) {
            fastDateFormat(c.getTimeInMillis(), false, writer);
        } else {
            format(N.asJUDate(c), format, timeZone, writer);
        }
    }

    public static String format(XMLGregorianCalendar c) {
        return format(c, null, null);
    }

    public static String format(XMLGregorianCalendar c, String format) {
        return format(c, format, null);
    }

    public static String format(XMLGregorianCalendar c, String format, TimeZone timeZone) {
        if ((format == null) && (timeZone == null)) {
            BufferedWriter cbuff = ObjectFactory.createBufferedWriter();
            fastDateFormat(c.toGregorianCalendar().getTimeInMillis(), false, cbuff);

            String st = cbuff.toString();
            ObjectFactory.recycle(cbuff);

            return st;
        }

        return format(N.asJUDate(c.toGregorianCalendar()), format, timeZone);
    }

    public static void format(XMLGregorianCalendar c, String format, TimeZone timeZone, Writer writer) {
        if ((format == null) && (timeZone == null)) {
            fastDateFormat(c.toGregorianCalendar().getTimeInMillis(), false, writer);
        } else {
            format(N.asJUDate(c.toGregorianCalendar()), format, timeZone, writer);
        }
    }

    private static void fastDateFormat(long timeInMillis, boolean isTimestamp, Writer writer) {
        Calendar c = utcCalendarPool.poll();

        if (c == null) {
            c = Calendar.getInstance(UTC_TIME_ZONE);
        }

        c.setTimeInMillis(timeInMillis);

        int year = c.get(Calendar.YEAR);
        int month = c.get(Calendar.MONTH) + 1;
        int day = c.get(Calendar.DAY_OF_MONTH);
        int hour = c.get(Calendar.HOUR_OF_DAY);
        int minute = c.get(Calendar.MINUTE);
        int second = c.get(Calendar.SECOND);

        try {
            writer.write(cbufOfSTDInt[4][year]);
            writer.write('-');
            writer.write(cbufOfSTDInt[2][month]);
            writer.write('-');
            writer.write(cbufOfSTDInt[2][day]);
            writer.write('T');
            writer.write(cbufOfSTDInt[2][hour]);
            writer.write(':');
            writer.write(cbufOfSTDInt[2][minute]);
            writer.write(':');
            writer.write(cbufOfSTDInt[2][second]);

            if (isTimestamp) {
                writer.write('.');

                int milliSecond = c.get(Calendar.MILLISECOND);
                writer.write(cbufOfSTDInt[3][milliSecond]);
            }

            writer.write('Z');
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            utcCalendarPool.add(c);
        }
    }

    public static String format(String format, double d) {
        final DecimalFormat numberFormat = new DecimalFormat(format);

        return numberFormat.format(d);
    }

    /**
     * Replace ''' or '"' with '\'' or '\"' if the previous char of the
     * quotation is not '\'. original String is returned if the specified String
     * is {@code null} or empty.
     *
     * @param str
     * @return
     */
    public static String formalizeQuotation(String str) {
        if (isNullOrEmpty(str)) {
            return str;
        }

        int length = str.length();
        StringBuilder sb = ObjectFactory.createStringBuilder();
        char c = 0;

        for (int i = 0; i < length; i++) {
            c = str.charAt(i);

            if ((c == '\\') && (i < (length - 1))) {
                sb.append(c);
                sb.append(str.charAt(++i));
            } else if ((c == '\'') || (c == '"')) {
                sb.append('\\');
                sb.append(c);
            } else {
                sb.append(c);
            }
        }

        String st = sb.toString();
        ObjectFactory.recycle(sb);

        return st;
    }

    public static boolean isPrimaryType(Class<?> cls) {
        return PRIMARY_TYPE.contains(cls);
    }

    public static Map<String, String> properties2Map(Properties props) {
        Map<String, String> m = N.asMap();

        for (Object key : props.keySet()) {
            m.put(N.toString(key), N.toString(props.get(key)));
        }

        return m;
    }

    /**
    * {@code true } is returned if {@code min < value < max}. Otherwise,
    * {@code false} is returned.
    *
    * @param value
    * @param min
    * @param max
    * @return
    */
    public static boolean isBetween(int value, int min, int max) {
        return (value > min) && (value < max);
    }

    public static <T> T createMask(Class<T> interfaceClass) {
        InvocationHandler h = new InvocationHandler() {
                @Override
                public Object invoke(Object proxy, Method method, Object[] args)
                    throws Throwable {
                    // TODO Auto-generated method stub
                    return null;
                }
            };

        return N.newProxyInstance(interfaceClass, h);
    }

    static int getClazzEnumOrder(Class<?> cls) {
        Integer order = cls.isArray() ? CLASS_ENUM_ORDER.get(cls.getComponentType()) : CLASS_ENUM_ORDER.get(cls);

        return (order == null) ? 0 : order.intValue();
    }

    static ClazzParserType getClazzParserType(Class<?> cls) {
        ClazzParserType clsParserType = clsParserTypePool.get(cls);

        if (clsParserType == null) {
            if (cls.isArray()) {
                clsParserType = ClazzParserType.ARRAY;
            } else if (Collection.class.isAssignableFrom(cls)) {
                clsParserType = ClazzParserType.COLLECTION;
            } else if (Map.class.isAssignableFrom(cls)) {
                clsParserType = ClazzParserType.MAP;
            } else if (getPropGetMethodList(cls).size() > 0) {
                clsParserType = ClazzParserType.ENTITY;
            } else {
                clsParserType = ClazzParserType.SINGLE_VALUE;
            }

            clsParserTypePool.put(cls, clsParserType);
        }

        return clsParserType;
    }

    static final void dummyMethod() {
    }

    private static final class DummyClass {
    }
    static enum ClazzParserType {ARRAY, COLLECTION, MAP, ENTITY, SINGLE_VALUE;
    }
}
